# Copyright (C) 2015 UCSC Computational Genomics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import

from collections import defaultdict
import os
import time
import pickle
from Queue import Queue
import logging
import sys

import mesos.interface
import mesos.native

from mesos.interface import mesos_pb2

from toil.batchSystems.abstractBatchSystem import AbstractBatchSystem
from toil.batchSystems.mesos import ToilJob, ResourceRequirement, TaskData

log = logging.getLogger(__name__)


class MesosBatchSystem(AbstractBatchSystem, mesos.interface.Scheduler):
    """
    A toil batch system implementation that uses Apache Mesos to distribute toil jobs as Mesos tasks over a
    cluster of slave nodes. A Mesos framework consists of a scheduler and an executor. This class acts as the
    scheduler and is typically run on the master node that also runs the Mesos master process with which the
    scheduler communicates via a driver component. The executor is implemented in a separate class. It is run on each
    slave node and communicates with the Mesos slave process via another driver object. The scheduler may also be run
    on a separate node from the master, which we then call somewhat ambiguously the driver node.
    """

    @staticmethod
    def supportsHotDeployment():
        return True

    def __init__(self, config, maxCores, maxMemory, maxDisk, masterIP,
                 userScript=None, toilDistribution=None):
        AbstractBatchSystem.__init__(self, config, maxCores, maxMemory, maxDisk)
        # The hot-deployed resources representing the user script and the toil distribution
        # respectively. Will be passed along in every Mesos task. See
        # toil.common.HotDeployedResource for details.
        self.userScript = userScript
        self.toilDistribution = toilDistribution

        # Written to when mesos kills tasks, as directed by toil
        self.killedSet = set()

        # Dictionary of queues, which toil assigns jobs to. Each queue represents a job type,
        # defined by resource usage
        self.jobQueueList = defaultdict(list)

        # IP of mesos master. specified in MesosBatchSystem, currently loopback
        self.masterIP = masterIP

        # queue of jobs to kill, by jobID.
        self.killSet = set()

        # Dict of launched jobIDs to TaskData named tuple. Contains start time, executorID, and slaveID.
        self.runningJobMap = {}

        # Queue of jobs whose status has been updated, according to mesos. Req'd by toil
        self.updatedJobsQueue = Queue()

        # Wether to use implicit/explicit acknowledgments
        self.implicitAcknowledgements = self.getImplicit()

        # Reference to the Mesos driver used by this scheduler, to be instantiated in run()
        self.driver = None

        # FIXME: This comment makes no sense to me

        # Returns Mesos executor object, which is merged into Mesos tasks as they are built
        self.executor = self.buildExecutor()

        self.nextJobID = 0
        self.lastReconciliation = time.time()
        self.reconciliationPeriod = 120

        # Start the driver
        self._startDriver()

    def issueBatchJob(self, command, memory, cores, disk):
        """
        Issues the following command returning a unique jobID. Command is the string to run, memory is an int giving
        the number of bytes the job needs to run in and cores is the number of cpus needed for the job and error-file
        is the path of the file to place any std-err/std-out in.
        """
        # puts job into job_type_queue to be run by Mesos, AND puts jobID in current_job[]
        self.checkResourceRequest(memory, cores, disk)
        jobID = self.nextJobID
        self.nextJobID += 1

        job = ToilJob(jobID=jobID,
                         resources=ResourceRequirement(memory=memory, cores=cores, disk=disk),
                         command=command,
                         userScript=self.userScript,
                         toilDistribution=self.toilDistribution)
        job_type = job.resources

        log.debug("Queueing the job command: %s with job id: %s ..." % (command, str(jobID)))
        self.jobQueueList[job_type].append(job)
        log.debug("... queued")

        return jobID

    def killBatchJobs(self, jobIDs):
        """
        Kills the given job IDs.
        """
        localSet = set()
        if self.driver is None:
            raise RuntimeError("There is no scheduler driver")
        for jobID in jobIDs:
            log.debug("passing tasks to kill to Mesos driver")
            self.killSet.add(jobID)
            localSet.add(jobID)

            if jobID not in self.getIssuedBatchJobIDs():
                self.killSet.remove(jobID)
                localSet.remove(jobID)
                log.debug("Job %s already finished", jobID)
            else:
                taskId = mesos_pb2.TaskID()
                taskId.value = str(jobID)
                self.driver.killTask(taskId)

        while localSet:
            log.debug("in while loop")
            intersection = localSet.intersection(self.killedSet)
            localSet -= intersection
            self.killedSet -= intersection
            if not intersection:
                log.debug("sleeping in the while")
                time.sleep(1)

    def getIssuedBatchJobIDs(self):
        """
        A list of jobs (as jobIDs) currently issued (may be running, or maybe just waiting).
        """
        # TODO: Ensure jobList holds jobs that have been "launched" from Mesos
        jobList = []
        for k, queue in self.jobQueueList.iteritems():
            for item in queue:
                jobList.append(item.jobID)
        for k, v in self.runningJobMap.iteritems():
            jobList.append(k)

        return jobList

    def getRunningBatchJobIDs(self):
        """
        Gets a map of jobs (as jobIDs) currently running (not just waiting) and a how long they have been running for
        (in seconds).
        """
        currentTime = dict()
        for jobID, data in self.runningJobMap.iteritems():
            currentTime[jobID] = time.time() - data.startTime
        return currentTime

    def getUpdatedBatchJob(self, maxWait):
        """
        Gets a job that has updated its status, according to the job manager. Max wait gives the number of seconds to
        pause waiting for a result. If a result is available returns (jobID, exitValue) else it returns None.
        """
        i = self.getFromQueueSafely(self.updatedJobsQueue, maxWait)
        if i is None:
            return None
        jobID, retcode = i
        self.updatedJobsQueue.task_done()
        log.debug("Job updated with code {}".format(retcode))
        return i

    def getWaitDuration(self):
        """
        Gets the period of time to wait (floating point, in seconds) between checking for
        missing/overlong jobs.
        """
        return self.reconciliationPeriod

    @classmethod
    def getRescueBatchJobFrequency(cls):
        """
        Parasol leaks jobs, but rescuing jobs involves calls to parasol list jobs and pstat2,
        making it expensive. We allow this every 10 minutes..
        """
        return 1800  # Half an hour

    def buildExecutor(self):
        """
        Creates and returns an ExecutorInfo instance representing our executor implementation.
        """

        def scriptPath(executorClass):
            path = sys.modules[executorClass.__module__].__file__
            if path.endswith('.pyc'):
                path = path[:-1]
            return path

        executorInfo = mesos_pb2.ExecutorInfo()
        # The executor program is installed as a setuptools entry point by setup.py
        executorInfo.command.value = "toil-mesos-executor"
        executorInfo.executor_id.value = "toilExecutor"
        executorInfo.name = "Test Executor (Python)"
        executorInfo.source = "python_test"
        return executorInfo

    def getImplicit(self):
        """
        Determine whether to run with implicit or explicit acknowledgements.
        """
        implicitAcknowledgements = 1
        if os.getenv("MESOS_EXPLICIT_ACKNOWLEDGEMENTS"):
            log.debug("Enabling explicit status update acknowledgements")
            implicitAcknowledgements = 0

        return implicitAcknowledgements

    def _startDriver(self):
        """
        The Mesos driver thread which handles the scheduler's communication with the Mesos master
        """
        framework = mesos_pb2.FrameworkInfo()
        framework.user = ""  # Have Mesos fill in the current user.
        framework.name = "toil"


        if os.getenv("MESOS_CHECKPOINT"):
            log.debug("Enabling checkpoint for the framework")
            framework.checkpoint = True

        if os.getenv("MESOS_AUTHENTICATE"):
            raise NotImplementedError("Authentication is currently not supported")
        else:
            framework.principal = framework.name
            self.driver = mesos.native.MesosSchedulerDriver(self, framework, self.masterIP,
                                                            self.implicitAcknowledgements)
        assert self.driver.start() == mesos_pb2.DRIVER_RUNNING

    def shutdown(self):
        log.info("Stopping Mesos driver")
        self.driver.stop()
        log.info("Joining Mesos driver")
        driver_result = self.driver.join()
        log.info("Joined Mesos driver")
        if driver_result != mesos_pb2.DRIVER_STOPPED:
            raise RuntimeError("Mesos driver failed with %i", driver_result)

    def registered(self, driver, frameworkId, masterInfo):
        """
        Invoked when the scheduler successfully registers with a Mesos master
        """
        log.debug("Registered with framework ID %s" % frameworkId.value)

    def _sortJobsByResourceReq(self):
        job_types = list(self.jobQueueList.keys())
        # sorts from largest to smallest core usage
        # TODO: add a size() method to ResourceSummary and use it as the key. Ask me why.
        job_types.sort(key=lambda resourceRequirement: ResourceRequirement.cores)
        job_types.reverse()
        return job_types

    def _declineAllOffers(self, driver, offers):
        for offer in offers:
            log.debug("No jobs to assign. Rejecting offer".format(offer.id.value))
            driver.declineOffer(offer.id)

    def _determineOfferResources(self, offer):
        offerCores = 0
        offerMem = 0
        offerStor = 0
        for resource in offer.resources:
            if resource.name == "cpus":
                offerCores += resource.scalar.value
            elif resource.name == "mem":
                offerMem += resource.scalar.value
            elif resource.name == "disk":
                offerStor += resource.scalar.value
        return offerCores, offerMem, offerStor

    def _prepareToRun(self, job_type, offer, index):
        jt_job = self.jobQueueList[job_type][index]  # get the first element to insure FIFO
        task = self._createTask(jt_job, offer)
        return task

    def _deleteByJobID(self, jobID, ):
        # FIXME: not efficient, I'm sure.
        for key, jobType in self.jobQueueList.iteritems():
            for job in jobType:
                if jobID == job.jobID:
                    jobType.remove(job)

    def _updateStateToRunning(self, offer, task):
        self.runningJobMap[int(task.task_id.value)] = TaskData(startTime=time.time(),
                                                               slaveID=offer.slave_id,
                                                               executorID=task.executor.executor_id)
        self._deleteByJobID(int(task.task_id.value))

    def resourceOffers(self, driver, offers):
        """
        Invoked when resources have been offered to this framework.
        """
        job_types = self._sortJobsByResourceReq()

        if len(job_types) == 0 or (len(self.getIssuedBatchJobIDs()) - len(self.getRunningBatchJobIDs()) == 0):
            log.debug("Declining offers")
            # If there are no jobs, we can get stuck with no jobs and no new offers until we decline it.
            self._declineAllOffers(driver, offers)
            return

        # Right now, gives priority to largest jobs
        for offer in offers:
            tasks = []
            # TODO: In an offer, can there ever be more than one resource with the same name?
            offerCores, offerMem, offerStor = self._determineOfferResources(offer)
            log.debug("Received offer %s with cores: %s, disk: %s, and mem: %s" \
                      % (offer.id.value, offerCores, offerStor, offerMem))
            remainingCores = offerCores
            remainingMem = offerMem
            remainingStor = offerStor

            for job_type in job_types:
                nextToLaunchIndex = 0
                # Because we are not removing from the list until outside of the while loop, we must decrement the
                # number of jobs left to run ourselves to avoid infinite loop.
                while (len(self.jobQueueList[job_type]) - nextToLaunchIndex > 0) and \
                                remainingCores >= job_type.cores and \
                                remainingStor >= self.__bytesToMB(job_type.disk) and \
                                remainingMem >= self.__bytesToMB(job_type.memory):  # toil specifies mem in bytes.

                    task = self._prepareToRun(job_type, offer, nextToLaunchIndex)
                    if int(task.task_id.value) not in self.runningJobMap:
                        # check to make sure task isn't already running (possibly in very unlikely edge case)
                        tasks.append(task)
                        log.info("Preparing to launch Mesos task %s using offer %s..." % (
                            task.task_id.value, offer.id.value))
                        remainingCores -= job_type.cores
                        remainingMem -= self.__bytesToMB(job_type.memory)
                        remainingStor -= job_type.disk
                    nextToLaunchIndex += 1

            # If we put the launch call inside the while loop, multiple accepts are used on the same offer.
            driver.launchTasks(offer.id, tasks)

            for task in tasks:
                self._updateStateToRunning(offer, task)
                log.info("...launching Mesos task %s" % task.task_id.value)

            if len(tasks) == 0:
                log.debug("Offer %s not large enough to run any tasks. Required: %s Offered: %s"
                          % (offer.id.value, job_types[-1], (self.__mbToBytes(offerMem), offerCores, self.__mbToBytes(offerStor))))

    def _createTask(self, jt_job, offer):
        """
        Build the Mesos task object from the toil job here to avoid further cluttering resourceOffers
        """
        task = mesos_pb2.TaskInfo()
        task.task_id.value = str(jt_job.jobID)
        task.slave_id.value = offer.slave_id.value
        task.name = "task %d" % jt_job.jobID

        # assigns toil command to task
        task.data = pickle.dumps(jt_job)

        task.executor.MergeFrom(self.executor)

        cpus = task.resources.add()
        cpus.name = "cpus"
        cpus.type = mesos_pb2.Value.SCALAR
        cpus.scalar.value = jt_job.resources.cores

        disk = task.resources.add()
        disk.name = "disk"
        disk.type = mesos_pb2.Value.SCALAR
        if self.__bytesToMB(jt_job.resources.disk) > 1:
            disk.scalar.value = self.__bytesToMB(jt_job.resources.disk)
        else:
            log.warning("Job %s uses less disk than mesos requires. Rounding %s bytes up to 1 mb" %
                        (jt_job.jobID, jt_job.resources.disk))
            disk.scalar.value = 1
        mem = task.resources.add()
        mem.name = "mem"
        mem.type = mesos_pb2.Value.SCALAR
        if self.__bytesToMB(jt_job.resources.memory) > 1:
            mem.scalar.value = self.__bytesToMB(jt_job.resources.memory)
        else:
            log.warning("Job %s uses less memory than mesos requires. Rounding %s bytes up to 1 mb" %
                        (jt_job.jobID, jt_job.resources.memory))
            mem.scalar.value = 1
        return task

    def __updateState(self, intID, exitStatus):
        self.updatedJobsQueue.put((intID, exitStatus))
        del self.runningJobMap[intID]

    def statusUpdate(self, driver, update):
        """
        Invoked when the status of a task has changed (e.g., a slave is lost and so the task is lost, a task finishes
        and an executor sends a status update saying so, etc). Note that returning from this callback _acknowledges_
        receipt of this status update! If for whatever reason the scheduler aborts during this callback (or the
        process exits) another status update will be delivered (note, however, that this is currently not true if the
        slave sending the status update is lost/fails during that time).
        """
        log.debug("Task %s is in a state %s" % \
                  (update.task_id.value, mesos_pb2.TaskState.Name(update.state)))

        intID = int(update.task_id.value)  # toil keeps jobIds as ints
        stringID = update.task_id.value  # Mesos keeps jobIds as strings

        try:
            self.killSet.remove(intID)
        except KeyError:
            pass
        else:
            self.killedSet.add(intID)

        if update.state == mesos_pb2.TASK_FINISHED:
            self.__updateState(intID, 0)

        if update.state == mesos_pb2.TASK_LOST or \
                        update.state == mesos_pb2.TASK_FAILED or \
                        update.state == mesos_pb2.TASK_KILLED or \
                        update.state == mesos_pb2.TASK_ERROR:
            log.warning("Task %s is in unexpected state %s with message '%s'" \
                        % (stringID, mesos_pb2.TaskState.Name(update.state), update.message))
            self.__updateState(intID, 1)

        # Explicitly acknowledge the update if implicit acknowledgements are not being used.
        if not self.implicitAcknowledgements:
            driver.acknowledgeStatusUpdate(update)

    def frameworkMessage(self, driver, executorId, slaveId, message):
        """
        Invoked when an executor sends a message.
        """
        log.debug("Executor {} on slave {} sent a message: {}".format(executorId, slaveId, message))

    def __reconcile(self, driver):
        """
        Queries the master about a list of running tasks. If the master has no knowledge of them, their state will be
        updated to LOST.
        """
        # FIXME: we need additional reconciliation. What about the tasks the master knows about but haven't updated?
        now = time.time()
        if now > self.lastReconciliation + self.reconciliationPeriod:
            self.lastReconciliation = now
            driver.reconcileTasks(list(self.runningJobMap.keys()))

    def reregistered(self, driver, masterInfo):
        """
        Invoked when the scheduler re-registers with a newly elected Mesos master.
        """
        log.debug("Registered with new master")

    def executorLost(self, driver, executorId, slaveId, status):
        """
        Invoked when an executor has exited/terminated.
        """
        log.warning("executor %s lost.".format(executorId))

    @staticmethod
    def __bytesToMB(mem):
        """
        used when converting toil reqs to Mesos reqs
        """
        return mem / 1024 / 1024

    @staticmethod
    def __mbToBytes(mem):
        """
        used when converting Mesos reqs to Toil reqs
        """
        return mem * 1024 * 1024



# Copyright (C) 2015 UCSC Computational Genomics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import
from contextlib import contextmanager
import logging
import marshal as pickler
#import cPickle as pickler
#import pickle as pickler
#import json as pickler    
import random
import shutil
import os
import tempfile
import stat
from toil.lib.bioio import absSymPath
from toil.jobStores.abstractJobStore import AbstractJobStore, NoSuchJobException, \
    NoSuchFileException
from toil.jobWrapper import JobWrapper

logger = logging.getLogger( __name__ )

class FileJobStore(AbstractJobStore):
    """Represents the toil using a network file system. For doc-strings
    of functions see AbstractJobStore.
    """

    def __init__(self, jobStoreDir, config=None):
        """
        :param jobStoreDir: Place to create jobStore
        :param config: See jobStores.abstractJobStore.AbstractJobStore.__init__
        :raise RuntimeError: if config != None and the jobStore already exists or
        config == None and the jobStore does not already exists. 
        """
        #This is root directory in which everything in the store is kept
        self.jobStoreDir = absSymPath(jobStoreDir)
        logger.info("Jobstore directory is: %s", self.jobStoreDir)
        #Safety checks for existing jobStore
        self._checkJobStoreCreation(config != None, os.path.exists(self.jobStoreDir), self.jobStoreDir)
        #Directory where temporary files go
        self.tempFilesDir = os.path.join(self.jobStoreDir, "tmp")
        #Creation of jobStore, if necessary
        if config != None:
            os.mkdir(self.jobStoreDir)
            os.mkdir(self.tempFilesDir)  
        #Parameters for creating temporary files
        self.validDirs = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789"
        self.levels = 2
        super( FileJobStore, self ).__init__( config=config )
        
    def deleteJobStore(self):
        if os.path.exists(self.jobStoreDir):
            shutil.rmtree(self.jobStoreDir)
    
    ##########################################
    #The following methods deal with creating/loading/updating/writing/checking for the
    #existence of jobs
    ########################################## 
    
    def create(self, command, memory, cores, disk, updateID=None,
               predecessorNumber=0):
        #The absolute path to the job directory.
        absJobDir = tempfile.mkdtemp(prefix="job", dir=self._getTempSharedDir())
        #Sub directory to put temporary files associated with the job in
        os.mkdir(os.path.join(absJobDir, "g"))
        #Make the job
        job = JobWrapper(command=command, memory=memory, cores=cores, disk=disk,
                  jobStoreID=self._getRelativePath(absJobDir), 
                  remainingRetryCount=self._defaultTryCount( ), 
                  updateID=updateID,
                  predecessorNumber=predecessorNumber)
        #Write job file to disk
        self.update(job)
        return job
    
    def exists(self, jobStoreID):
        return os.path.exists(self._getJobFileName(jobStoreID))
    
    def getPublicUrl( self,  jobStoreFileID):
        self._checkJobStoreFileID(jobStoreFileID)
        jobStorePath = self._getAbsPath(jobStoreFileID)
        if os.path.exists(jobStorePath):
            return 'file:'+jobStorePath
        else:
            raise NoSuchFileException(jobStoreFileID)

    def getSharedPublicUrl( self,  FileName):
        jobStorePath = self.jobStoreDir+'/'+FileName
        if os.path.exists(jobStorePath):
            return 'file:'+jobStorePath
        else:
            raise NoSuchFileException(FileName)

    def load(self, jobStoreID):
        self._checkJobStoreId(jobStoreID)
        #Load a valid version of the job
        jobFile = self._getJobFileName(jobStoreID)
        with open(jobFile, 'r') as fileHandle:
            job = JobWrapper.fromDict(pickler.load(fileHandle))
        #The following cleans up any issues resulting from the failure of the 
        #job during writing by the batch system.
        if os.path.isfile(jobFile + ".new"):
            logger.warn("There was a .new file for the job: %s", jobStoreID)
            os.remove(jobFile + ".new")
            job.setupJobAfterFailure(self.config)
        return job
    
    def update(self, job):
        #The job is serialised to a file suffixed by ".new"
        #The file is then moved to its correct path.
        #Atomicity guarantees use the fact the underlying file systems "move"
        #function is atomic. 
        with open(self._getJobFileName(job.jobStoreID) + ".new", 'w') as f:
            pickler.dump(job.toDict(), f)
        #This should be atomic for the file system
        os.rename(self._getJobFileName(job.jobStoreID) + ".new", self._getJobFileName(job.jobStoreID))
    
    def delete(self, jobStoreID):
        #The jobStoreID is the relative path to the directory containing the job,
        #removing this directory deletes the job.
        if self.exists(jobStoreID):
            shutil.rmtree(self._getAbsPath(jobStoreID))
 
    def jobs(self):
        #Walk through list of temporary directories searching for jobs
        for tempDir in self._tempDirectories():
            for i in os.listdir(tempDir):
                if i.startswith( 'job' ):
                    yield self.load(self._getRelativePath(os.path.join(tempDir, i)))
 
    ##########################################
    #Functions that deal with temporary files associated with jobs
    ##########################################    
    
    def writeFile(self, jobStoreID, localFilePath):
        self._checkJobStoreId(jobStoreID)
        fd, absPath = self._getJobTempFile(jobStoreID)
        shutil.copyfile(localFilePath, absPath)
        os.close(fd)
        return self._getRelativePath(absPath)

    def updateFile(self, jobStoreFileID, localFilePath):
        self._checkJobStoreFileID(jobStoreFileID)
        shutil.copyfile(localFilePath, self._getAbsPath(jobStoreFileID))
    
    def readFile(self, jobStoreFileID, localFilePath):
        self._checkJobStoreFileID(jobStoreFileID)
        shutil.copyfile(self._getAbsPath(jobStoreFileID), localFilePath)
    
    def deleteFile(self, jobStoreFileID):
        if not self.fileExists(jobStoreFileID):
            return
        os.remove(self._getAbsPath(jobStoreFileID))
        
    def fileExists(self, jobStoreFileID):
        absPath = self._getAbsPath(jobStoreFileID)
        try:
            st = os.stat(absPath)
        except os.error:
            return False
        if not stat.S_ISREG(st.st_mode):
            raise NoSuchFileException("Path %s is not a file in the jobStore" % jobStoreFileID)
        return True
    
    @contextmanager
    def writeFileStream(self, jobStoreID):
        self._checkJobStoreId(jobStoreID)
        fd, absPath =  self._getJobTempFile(jobStoreID)
        with open(absPath, 'w') as f:
            yield f, self._getRelativePath(absPath)
        os.close(fd) #Close the os level file descript

    @contextmanager
    def updateFileStream(self, jobStoreFileID):
        self._checkJobStoreFileID(jobStoreFileID)
        # File objects are context managers (CM) so we could simply return what open returns.
        # However, it is better to wrap it in another CM so as to prevent users from accessing
        # the file object directly, without a with statement.
        with open(self._getAbsPath(jobStoreFileID), 'w') as f:
            yield f
    
    def getEmptyFileStoreID(self, jobStoreID):
        with self.writeFileStream(jobStoreID) as ( fileHandle, jobStoreFileID ):
            return jobStoreFileID
    
    @contextmanager
    def readFileStream(self, jobStoreFileID):
        self._checkJobStoreFileID(jobStoreFileID)
        with open(self._getAbsPath(jobStoreFileID), 'r') as f:
            yield f
            
    ##########################################
    #The following methods deal with shared files, i.e. files not associated 
    #with specific jobs.
    ##########################################  

    @contextmanager
    def writeSharedFileStream(self, sharedFileName, isProtected=True):
        # the isProtected parameter has no effect on the fileStore, but is needed on the awsJobStore
        assert self._validateSharedFileName( sharedFileName )
        with open( os.path.join( self.jobStoreDir, sharedFileName ), 'w' ) as f:
            yield f

    @contextmanager
    def readSharedFileStream(self, sharedFileName, isProtected=True):
        # the isProtected parameter has no effect on the fileStore, but is needed on the awsJobStore
        assert self._validateSharedFileName( sharedFileName )
        with open(os.path.join(self.jobStoreDir, sharedFileName), 'r') as f:
            yield f
             
    def writeStatsAndLogging(self, statsAndLoggingString):
        #Temporary files are placed in the set of temporary files/directoies
        fd, tempStatsFile = tempfile.mkstemp(prefix="stats", suffix=".new", dir=self._getTempSharedDir())
        with open(tempStatsFile, "w") as f:
            f.write(statsAndLoggingString)
        os.close(fd)
        os.rename(tempStatsFile, tempStatsFile[:-4]) #This operation is atomic
        
    def readStatsAndLogging( self, statsAndLoggingCallBackFn):
        numberOfFilesProcessed = 0
        for tempDir in self._tempDirectories():
            for tempFile in os.listdir(tempDir):
                if tempFile.startswith( 'stats' ):
                    absTempFile = os.path.join(tempDir, tempFile)
                    if not tempFile.endswith( '.new' ):
                        with open(absTempFile, 'r') as fH:
                            statsAndLoggingCallBackFn(fH)
                        numberOfFilesProcessed += 1
                        os.remove(absTempFile)
        return numberOfFilesProcessed
    
    ##########################################
    #Private methods
    ##########################################   
        
    def _getAbsPath(self, relativePath):
        """
        :rtype : string, string is the absolute path to a file path relative
        to the self.tempFilesDir.
        """
        return os.path.join(self.tempFilesDir, relativePath)
    
    def _getRelativePath(self, absPath):
        """
        absPath  is the absolute path to a file in the store,.
        
        :rtype : string, string is the path to the absPath file relative to the 
        self.tempFilesDir
        
        """
        return absPath[len(self.tempFilesDir)+1:]
    
    def _getJobFileName(self, jobStoreID):
        """
        :rtype : string, string is the file containing the serialised JobWrapper instance
        for the given job.
        """
        return os.path.join(self._getAbsPath(jobStoreID), "job")

    def _getJobTempFile(self, jobStoreID):
        """
        :rtype : file-descriptor, string, string is absolute path to a temporary file within
        the given job's (referenced by jobStoreID's) temporary file directory. The file-descriptor
        is integer pointing to open operating system file handle. Should be closed using os.close()
        after writing some material to the file.
        """
        fD, absPath = tempfile.mkstemp(suffix=".tmp", 
                                dir=os.path.join(self._getAbsPath(jobStoreID), "g"))
        return fD, absPath
    
    def _checkJobStoreId(self, jobStoreID):
        """
        Raises a NoSuchJobException if the jobStoreID does not exist.
        """
        if not self.exists(jobStoreID):
            raise NoSuchJobException(jobStoreID)
    
    def _checkJobStoreFileID(self, jobStoreFileID):
        """
        Raises NoSuchFileException if the jobStoreFileID does not exist or is not a file.
        """
        if not self.fileExists(jobStoreFileID):
            raise NoSuchFileException("File %s does not exist in jobStore" % jobStoreFileID)
    
    def _getTempSharedDir(self):
        """
        Gets a temporary directory in the hierarchy of directories in self.tempFilesDir.
        This directory may contain multiple shared jobs/files.
        
        :rtype : string, path to temporary directory in which to place files/directories.
        """
        tempDir = self.tempFilesDir
        for i in xrange(self.levels):
            tempDir = os.path.join(tempDir, random.choice(self.validDirs))
            if not os.path.exists(tempDir):
                try:
                    os.mkdir(tempDir)
                except os.error:
                    if not os.path.exists(tempDir): #In the case that a collision occurs and
                        #it is created while we wait then we ignore
                        raise
        return tempDir
     
    def _tempDirectories(self):
        """
        :rtype : an iterator to the temporary directories containing jobs/stats files
        in the hierarchy of directories in self.tempFilesDir
        """
        def _dirs(path, levels):
            if levels > 0:
                for subPath in os.listdir(path):
                    for i in _dirs(os.path.join(path, subPath), levels-1):
                        yield i
            else:
                yield path
        for tempDir in _dirs(self.tempFilesDir, self.levels):
            yield tempDir



#!/usr/bin/env python

# Copyright (C) 2015 UCSC Computational Genomics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
import unittest
import os
from toil.lib.bioio import system
from argparse import ArgumentParser
from toil.common import setupToil
from toil.job import Job
from toil.test import ToilTest
from toil.jobWrapper import JobWrapper

class JobWrapperTest(ToilTest):
    
    def setUp(self):
        super( JobWrapperTest, self ).setUp( )
        self.jobStorePath = self._getTestJobStorePath()
        parser = ArgumentParser()
        Job.Runner.addToilOptions(parser)
        options = parser.parse_args(args=[self.jobStorePath])
        self.contextManager = setupToil(options)
        config, batchSystem, self.jobStore = self.contextManager.__enter__()

    def tearDown(self):
        self.contextManager.__exit__(None, None, None)
        self.jobStore.deleteJobStore()
        self.assertFalse(os.path.exists(self.jobStorePath))
        super( JobWrapperTest, self ).tearDown( )
    
    def testJob(self):       
        """
        Tests functions of a job.
        """ 
    
        command = "by your command"
        memory = 2^32
        disk = 2^32
        cores = 1
        jobStoreID = 100
        remainingRetryCount = 5
        predecessorNumber = 0
        updateID = 1000
        
        j = JobWrapper(command, memory, cores, disk, jobStoreID, remainingRetryCount,
                  updateID, predecessorNumber)
        
        #Check attributes
        #
        self.assertEquals(j.command, command)
        self.assertEquals(j.memory, memory)
        self.assertEquals(j.disk, disk)
        self.assertEquals(j.cores, cores)
        self.assertEquals(j.jobStoreID, jobStoreID)
        self.assertEquals(j.remainingRetryCount, remainingRetryCount)
        self.assertEquals(j.predecessorNumber, predecessorNumber)
        self.assertEquals(j.updateID, updateID)
        self.assertEquals(j.stack, [])
        self.assertEquals(j.predecessorsFinished, set())
        self.assertEquals(j.logJobStoreFileID, None)
        
        #Check equals function
        j2 = JobWrapper(command, memory, cores, disk, jobStoreID, remainingRetryCount,
                  updateID, predecessorNumber)
        self.assertEquals(j, j2)
        #Change an attribute and check not equal
        j.predecessorsFinished = set(("1", "2"))
        self.assertNotEquals(j, j2)
        
        ###TODO test other functionality

if __name__ == '__main__':
    unittest.main()



# Copyright (C) 2015 UCSC Computational Genomics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import
import unittest
import os
from subprocess import CalledProcessError

from toil.lib.bioio import system
from toil.lib.bioio import getTempFile
from toil.test.sort.sortTest import makeFileToSort, checkEqual
from toil.common import toilPackageDirPath
from toil.test import ToilTest


class UtilsTest(ToilTest):
    """
    Tests the utilities that toil ships with, e.g. stats and status, in conjunction with restart
    functionality.
    """

    def setUp(self):
        super(UtilsTest, self).setUp()
        self.tempDir = self._createTempDir()
        self.tempFile = getTempFile(rootDir=self.tempDir)
        self.outputFile = getTempFile(rootDir=self.tempDir)
        self.toilDir = os.path.join(self.tempDir, "jobstore")
        self.assertFalse(os.path.exists(self.toilDir))
        self.lines = 1000
        self.maxLineLength = 10
        self.N = 1000
        makeFileToSort(self.tempFile, self.lines, self.maxLineLength)
        # First make our own sorted version
        with open(self.tempFile, 'r') as fileHandle:
            self.correctSort = fileHandle.readlines()
            self.correctSort.sort()

    def tearDown(self):
        ToilTest.tearDown(self)
        system("rm -rf %s" % self.tempDir)

    @property
    def toilMain(self):
        return self._getUtilScriptPath('toilMain')

    @property
    def sort(self):
        return os.path.join(toilPackageDirPath(), "test", "sort", "sort.py")

    def testUtilsSort(self):
        """
        Tests the restart, status and stats commands of the toil command line utility using the
        sort example.
        """
        # Get the sort command to run
        toilCommandString = ("{self.sort} "
                             "{self.toilDir} "
                             "--logLevel=DEBUG "
                             "--fileToSort={self.tempFile} "
                             "--N {self.N} --stats "
                             "--retryCount 2".format(**locals()))

        # Try restarting it to check that a JobStoreException is thrown
        self.assertRaises(CalledProcessError, system, toilCommandString + " --restart")
        # Check that trying to run it in restart mode does not create the jobStore
        self.assertFalse(os.path.exists(self.toilDir))

        # Status command
        toilStatusCommandString = ("{self.toilMain} status "
                                   "{self.toilDir} "
                                   "--failIfNotComplete".format(**locals()))

        # Run the script for the first time
        try:
            system(toilCommandString)
            finished = True
        except CalledProcessError:  # This happens when the script fails due to having unfinished jobs
            self.assertRaises(CalledProcessError, system, toilStatusCommandString)
            finished = False
        self.assertTrue(os.path.exists(self.toilDir))

        # Try running it without restart and check an exception is thrown
        self.assertRaises(CalledProcessError, system, toilCommandString)

        # Now restart it until done
        while not finished:
            try:
                system(toilCommandString + " --restart")
                finished = True
            except CalledProcessError:  # This happens when the script fails due to having unfinished jobs
                self.assertRaises(CalledProcessError, system, toilStatusCommandString)

        # Check the toil status command does not issue an exception
        system(toilStatusCommandString)

        # Check if we try to launch after its finished that we get a JobException
        self.assertRaises(CalledProcessError, system, toilCommandString + " --restart")

        # Check we can run 'toil stats'
        toilStatsString = ("{self.toilMain} stats "
                           "{self.toilDir} "
                           "--pretty".format(**locals()))
        system(toilStatsString)

        # Check the file is properly sorted
        with open(self.tempFile, 'r') as fileHandle:
            l2 = fileHandle.readlines()
            checkEqual(self.correctSort, l2)

        # Check we can run 'toil clean'
        toilCleanString = ("{self.toilMain} clean "
                           "{self.toilDir}".format(**locals()))
        system(toilCleanString)
            
    def testUtilsStatsSort(self):
        """
        Tests the stats commands on a complete run of the stats test.
        """
        # Get the sort command to run
        toilCommandString = ("{self.sort} "
                             "{self.toilDir} "
                             "--logLevel=DEBUG "
                             "--fileToSort={self.tempFile} "
                             "--N {self.N} --stats "
                             "--retryCount 99".format(**locals()))

        # Run the script for the first time
        system(toilCommandString)
        self.assertTrue(os.path.exists(self.toilDir))

        # Check we can run 'toil stats'
        rootPath = os.path.join(toilPackageDirPath(), "utils")
        toilStatsString = ("{self.toilMain} stats "
                           "{self.toilDir} --pretty".format(**locals()))
        system(toilStatsString)

        # Check the file is properly sorted
        with open(self.tempFile, 'r') as fileHandle:
            l2 = fileHandle.readlines()
            checkEqual(self.correctSort, l2)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

# Copyright (C) 2015 UCSC Computational Genomics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
import logging
import multiprocessing
import os
import random
import subprocess
import time
import math
from threading import Thread
from threading import Semaphore, Lock, Condition
from Queue import Queue, Empty

from toil.batchSystems.abstractBatchSystem import AbstractBatchSystem

logger = logging.getLogger(__name__)


class SingleMachineBatchSystem(AbstractBatchSystem):
    """
    The interface for running jobs on a single machine, runs all the jobs you give it as they come in, but in parallel.
    """

    numCores = multiprocessing.cpu_count()

    def __init__(self, config, maxCores, maxMemory, maxDisk):
        assert type(maxCores) == int
        if maxCores > self.numCores:
            logger.warn('Limiting maxCores to CPU count of system (%i).', self.numCores)
            maxCores = self.numCores
        AbstractBatchSystem.__init__(self, config, maxCores, maxMemory, maxDisk)
        assert self.maxCores >= 1
        assert self.maxMemory >= 1
        # The scale allows the user to apply a factor to each task's cores requirement, thereby squeezing more tasks
        # onto each core (scale < 1) or stretching tasks over more cores (scale > 1).
        self.scale = config.scale
        # The minimal fractional CPU. Tasks with a smaller cores requirement will be rounded up to this value. One
        # important invariant of this class is that each worker thread represents a CPU requirement of minCores,
        # meaning that we can never run more than numCores / minCores jobs concurrently. With minCores set to .1,
        # a task with cores=1 will occupy 10 workers. One of these workers will be blocked on the Popen.wait() call for
        # the worker.py child process, the others will be blocked on the acquiring the core semaphore.
        self.minCores = 0.1
        # Number of worker threads that will be started
        self.numWorkers = int(self.maxCores / self.minCores)
        # A counter to generate job IDs and a lock to guard it
        self.jobIndex = 0
        self.jobIndexLock = Lock()
        # A dictionary mapping IDs of submitted jobs to those jobs
        self.jobs = {}
        # A queue of jobs waiting to be executed. Consumed by the workers.
        self.inputQueue = Queue()
        # A queue of finished jobs. Produced by the workers.
        self.outputQueue = Queue()
        # A dictionary mapping IDs of currently running jobs to their Info objects
        self.runningJobs = {}
        # The list of worker threads
        self.workerThreads = []
        # A semaphore representing available CPU in units of minCores
        self.coreSemaphore = Semaphore(self.numWorkers)
        # A counter representing failed acquisitions of the semaphore, also in units of minCores, and a lock to guard it
        self.coreOverflow = 0
        self.coreOverflowLock = Lock()
        # A lock to work around the lack of thread-safety in Python's subprocess module
        self.popenLock = Lock()
        # A counter representing available memory in bytes
        self.memoryPool = self.maxMemory
        # A condition object used to guard it (a semphore would force us to acquire each unit of memory individually)
        self.memoryCondition = Condition()
        logger.info('Setting up the thread pool with %i workers, '
                    'given a minimum CPU fraction of %f '
                    'and a maximum CPU value of %i.', self.numWorkers, self.minCores, maxCores)
        for i in xrange(self.numWorkers):
            worker = Thread(target=self.worker, args=(self.inputQueue,))
            self.workerThreads.append(worker)
            worker.start()

    # The input queue is passed as an argument because the corresponding attribute is reset to None in shutdown()

    def worker(self, inputQueue):
        while True:
            args = inputQueue.get()
            if args is None:
                logger.debug('Received queue sentinel.')
                break
            jobCommand, jobID,jobCores, jobMem, jobDisk = args
            try:
                numThreads = int(jobCores / self.minCores)
                logger.debug('Acquiring %i bytes of memory from pool of %i.', jobMem, self.memoryPool)
                self.memoryCondition.acquire()
                while jobMem > self.memoryPool:
                    logger.debug('Waiting for memory condition to change.')
                    self.memoryCondition.wait()
                    logger.debug('Memory condition changed.')
                self.memoryPool -= jobMem
                self.memoryCondition.release()

                try:
                    logger.debug('Attempting to acquire %i threads for %i cpus submitted', numThreads, jobCores)
                    numThreadsAcquired = 0
                    # Acquire first thread blockingly
                    logger.debug('Acquiring semaphore blockingly.')
                    self.coreSemaphore.acquire(blocking=True)
                    try:
                        numThreadsAcquired += 1
                        logger.debug('Semaphore acquired.')
                        while numThreadsAcquired < numThreads:
                            # Optimistically and non-blockingly acquire remaining threads. For every failed attempt
                            # to acquire a thread, atomically increment the overflow instead of the semaphore such
                            # any thread that later wants to release a thread, can do so into the overflow,
                            # thereby effectively surrendering that thread to this job and not into the semaphore.
                            # That way we get to start a job with as many threads as are available, and later grab
                            # more as they become available.
                            if not self.coreSemaphore.acquire(blocking=False):
                                with self.coreOverflowLock:
                                    self.coreOverflow += 1
                            numThreadsAcquired += 1

                        logger.info("Executing command: '%s'.", jobCommand)
                        with self.popenLock:
                            popen = subprocess.Popen(jobCommand, shell=True)
                        info = Info(time.time(), popen, kill_intended=False)
                        self.runningJobs[jobID] = info
                        try:
                            statusCode = popen.wait()
                            if 0 != statusCode:
                                if statusCode != -9 or not info.kill_intended:
                                    logger.error("Got exit code %i (indicating failure) from command '%s'.", statusCode, jobCommand )
                        finally:
                            self.runningJobs.pop(jobID)
                    finally:
                        logger.debug('Releasing %i threads.', numThreadsAcquired)

                        with self.coreOverflowLock:
                            if self.coreOverflow > 0:
                                if self.coreOverflow > numThreadsAcquired:
                                    self.coreOverflow -= numThreadsAcquired
                                    numThreadsAcquired = 0
                                else:
                                    numThreadsAcquired -= self.coreOverflow
                                    self.coreOverflow = 0
                        for i in xrange(numThreadsAcquired):
                            self.coreSemaphore.release()
                finally:
                    logger.debug('Releasing %i memory back to pool', jobMem)
                    self.memoryCondition.acquire()
                    self.memoryPool += jobMem
                    self.memoryCondition.notifyAll()
                    self.memoryCondition.release()
            finally:
                # noinspection PyProtectedMember
                value = self.coreSemaphore._Semaphore__value
                logger.debug('Finished job. CPU semaphore value (approximate): %i, overflow: %i', value, self.coreOverflow)
                self.outputQueue.put((jobID, 0))
        logger.debug('Exiting worker thread normally.')

    def issueBatchJob(self, command, memory, cores, disk):
        """
        Adds the command and resources to a queue to be run.
        """
        # Round cores to minCores and apply scale
        cores = math.ceil(cores * self.scale / self.minCores) * self.minCores
        assert cores <= self.maxCores, \
            'job is requesting {} cores, which is greater than {} available on the machine. Scale currently set ' \
            'to {} consider adjusting job or scale.'.format(cores, multiprocessing.cpu_count(), self.scale)
        assert cores >= self.minCores
        assert memory <= self.maxMemory, 'job requests {} mem, only {} total available.'.format(memory, self.maxMemory)

        self.checkResourceRequest(memory, cores, disk)
        logger.debug("Issuing the command: %s with memory: %i, cores: %i, disk: %i" % (command, memory, cores, disk))
        with self.jobIndexLock:
            jobID = self.jobIndex
            self.jobIndex += 1
        self.jobs[jobID] = command
        self.inputQueue.put((command, jobID, cores, memory, disk))
        return jobID

    def killBatchJobs(self, jobIDs):
        """
        Kills jobs by ID
        """
        logger.debug('Killing jobs: {}'.format(jobIDs))
        for id in jobIDs:
            if id in self.runningJobs:
                info = self.runningJobs[id]
                info.kill_intended = True
                os.kill(info.popen.pid, 9)
                while id in self.runningJobs:
                    pass

    def getIssuedBatchJobIDs(self):
        """
        Just returns all the jobs that have been run, but not yet returned as updated.
        """
        return self.jobs.keys()

    def getRunningBatchJobIDs(self):
        """
        Return empty map
        """
        currentJobs = {}
        for jobID, info in self.runningJobs.iteritems():
            startTime = info.time
            currentJobs[jobID] = time.time() - startTime
        return currentJobs

    def shutdown(self):
        """
        Cleanly terminate worker threads. Add sentinels to inputQueue equal to maxThreads. Join all worker threads.
        """
        # Remove reference to inputQueue (raises exception if inputQueue is used after method call)
        inputQueue = self.inputQueue
        self.inputQueue = None
        for i in xrange(self.numWorkers):
            inputQueue.put(None)

        for thread in self.workerThreads:
            thread.join()

    def getUpdatedBatchJob(self, maxWait):
        """
        Returns a map of the run jobs and the return value of their processes.
        """
        try:
            i = self.outputQueue.get(timeout=maxWait)
        except Empty:
            return None
        jobID, exitValue = i
        self.jobs.pop(jobID)
        logger.debug("Ran jobID: %s with exit value: %i" % (jobID, exitValue))
        self.outputQueue.task_done()
        return jobID, exitValue

    @classmethod
    def getRescueBatchJobFrequency(cls):
        """
        This should not really occur, wihtout an error. To exercise the system we allow it every 90 minutes.
        """
        return 5400


class Info(object):
    def __init__(self, time, popen, kill_intended):
        self.time = time
        self.popen = popen
        self.kill_intended = kill_intended



#!/usr/bin/env python

# Copyright (C) 2015 UCSC Computational Genomics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
import os
import sys

if __name__ == "__main__":
    # FIXME: Until we use setuptools entry points, this is the only way to avoid a conflict between our own resource.py
    # and Python's
    toilSrcDir = os.path.dirname(os.path.realpath(__file__))
    sys.path = [directory for directory in sys.path if not os.path.realpath(directory) == toilSrcDir]

import traceback
import time
import socket
import logging
import xml.etree.cElementTree as ET
import cPickle
import shutil

logger = logging.getLogger( __name__ )

def truncateFile(fileNameString, tooBig=50000):
    """
    Truncates a file that is bigger than tooBig bytes, leaving only the 
    last tooBig bytes in the file.
    """
    if os.path.getsize(fileNameString) > tooBig:
        fh = open(fileNameString, 'rb+')
        fh.seek(-tooBig, 2) 
        data = fh.read()
        fh.seek(0) # rewind
        fh.write(data)
        fh.truncate()
        fh.close()
    
def nextOpenDescriptor():
    """Gets the number of the next available file descriptor.
    """
    descriptor = os.open("/dev/null", os.O_RDONLY)
    os.close(descriptor)
    return descriptor
    
def main():
    ########################################## 
    #Import necessary modules 
    ##########################################
    
    # This is assuming that worker.py is at a path ending in "/toil/worker.py".
    sourcePath = os.path.dirname(os.path.dirname(__file__))
    if sourcePath not in sys.path:
        sys.path.append(sourcePath)
    
    #Now we can import all the necessary functions
    from toil.lib.bioio import setLogLevel
    from toil.lib.bioio import getTotalCpuTime
    from toil.lib.bioio import getTotalCpuTimeAndMemoryUsage
    from toil.lib.bioio import getTempDirectory
    from toil.lib.bioio import makePublicDir
    from toil.lib.bioio import system
    from toil.common import loadJobStore
    from toil.job import Job
    
    ########################################## 
    #Input args
    ##########################################
    
    jobStoreString = sys.argv[1]
    jobStoreID = sys.argv[2]
    
    ##########################################
    #Load the jobStore/config file
    ##########################################
    
    jobStore = loadJobStore(jobStoreString)
    config = jobStore.config

    ##########################################
    #Load the environment for the job
    ##########################################
    
    #First load the environment for the job.
    with jobStore.readSharedFileStream("environment.pickle") as fileHandle:
        environment = cPickle.load(fileHandle)
    for i in environment:
        if i not in ("TMPDIR", "TMP", "HOSTNAME", "HOSTTYPE"):
            os.environ[i] = environment[i]
    # sys.path is used by __import__ to find modules
    if "PYTHONPATH" in environment:
        for e in environment["PYTHONPATH"].split(':'):
            if e != '':
                sys.path.append(e)

    setLogLevel(config.logLevel)

    tempRootDir = config.workDir
    if tempRootDir is not None and not os.path.exists(tempRootDir):
        raise RuntimeError("The temporary directory specified by workDir: %s does not exist" % tempRootDir)

    ##########################################
    #Setup the temporary directories.
    ##########################################
        
    #Dir to put all the temp files in. If tempRootDir is None, tempdir looks at environment variables to determine
    # where to put the tempDir.
    localWorkerTempDir = getTempDirectory(tempRootDir)
    
    ##########################################
    #Setup the logging
    ##########################################

    #This is mildly tricky because we don't just want to
    #redirect stdout and stderr for this Python process; we want to redirect it
    #for this process and all children. Consequently, we can't just replace
    #sys.stdout and sys.stderr; we need to mess with the underlying OS-level
    #file descriptors. See <http://stackoverflow.com/a/11632982/402891>
    
    #When we start, standard input is file descriptor 0, standard output is
    #file descriptor 1, and standard error is file descriptor 2.

    #What file do we want to point FDs 1 and 2 to?    
    tempWorkerLogPath = os.path.join(localWorkerTempDir, "worker_log.txt")
    
    #Save the original stdout and stderr (by opening new file descriptors to the
    #same files)
    origStdOut = os.dup(1)
    origStdErr = os.dup(2)
    
    #Open the file to send stdout/stderr to.
    logFh = os.open(tempWorkerLogPath, os.O_WRONLY | os.O_CREAT | os.O_APPEND)

    #Replace standard output with a descriptor for the log file
    os.dup2(logFh, 1)
    
    #Replace standard error with a descriptor for the log file
    os.dup2(logFh, 2)
    
    #Since we only opened the file once, all the descriptors duped from the
    #original will share offset information, and won't clobber each others'
    #writes. See <http://stackoverflow.com/a/5284108/402891>. This shouldn't
    #matter, since O_APPEND seeks to the end of the file before every write, but
    #maybe there's something odd going on...
    
    #Close the descriptor we used to open the file
    os.close(logFh)
    
    for handler in list(logger.handlers): #Remove old handlers
        logger.removeHandler(handler)
    
    #Add the new handler. The sys.stderr stream has been redirected by swapping
    #the file descriptor out from under it.
    logger.addHandler(logging.StreamHandler(sys.stderr))

    debugging = logging.getLogger().isEnabledFor(logging.DEBUG)
    ##########################################
    #Worker log file trapped from here on in
    ##########################################

    workerFailed = False
    elementNode = ET.Element("worker")
    messageNode = ET.SubElement(elementNode, "messages")
    messages = []
    try:

        #Put a message at the top of the log, just to make sure it's working.
        print "---TOIL WORKER OUTPUT LOG---"
        sys.stdout.flush()
        
        #Log the number of open file descriptors so we can tell if we're leaking
        #them.
        logger.debug("Next available file descriptor: {}".format(
            nextOpenDescriptor()))
    
        ##########################################
        #Load the job
        ##########################################
        
        job = jobStore.load(jobStoreID)
        logger.debug("Parsed job")
        
        ##########################################
        #Cleanup from any earlier invocation of the job
        ##########################################
        
        if job.command == None:
            while len(job.stack) > 0:
                jobs = job.stack[-1]
                #If the jobs still exist they have not been run, so break
                if jobStore.exists(jobs[0][0]):
                    break
                #However, if they are gone then we can remove them from the stack.
                #This is the only way to flush successors that have previously been run
                #, as jobs are, as far as possible, read only in the leader.
                job.stack.pop()
                
                
        #This cleans the old log file which may 
        #have been left if the job is being retried after a job failure.
        if job.logJobStoreFileID != None:
            job.clearLogFile(jobStore)
    
        ##########################################
        #Setup the stats, if requested
        ##########################################
        
        if config.stats:
            startTime = time.time()
            startClock = getTotalCpuTime()

        startTime = time.time() 
        while True:
            ##########################################
            #Run the job, if there is one
            ##########################################
            
            if job.command != None:
                if job.command.startswith( "_toil " ):
                    #Make a temporary file directory for the job
                    localTempDir = makePublicDir(os.path.join(localWorkerTempDir, "localTempDir"))
                    
                    #Is a job command
                    messages = Job._loadJob(job.command, jobStore)._execute( jobWrapper=job,
                                    stats=elementNode if config.stats else None, localTempDir=localTempDir,
                                    jobStore=jobStore)
                    
                    #Remove the temporary file directory
                    shutil.rmtree(localTempDir)
    
                else: #Is another command (running outside of jobs may be deprecated)
                    system(job.command)
            else:
                #The command may be none, in which case
                #the job is just a shell ready to be deleted
                assert len(job.stack) == 0
                break
            
            ##########################################
            #Establish if we can run another job within the worker
            ##########################################
            
            #No more jobs to run so quit
            if len(job.stack) == 0:
                break
            
            #Get the next set of jobs to run
            jobs = job.stack[-1]
            assert len(jobs) > 0
            
            #If there are 2 or more jobs to run in parallel we quit
            if len(jobs) >= 2:
                logger.debug("No more jobs can run in series by this worker,"
                            " it's got %i children", len(jobs)-1)
                break
            
            #We check the requirements of the job to see if we can run it
            #within the current worker
            successorJobStoreID, successorMemory, successorCores, successorsDisk, successorPredecessorID = jobs[0]
            if successorMemory > job.memory:
                logger.debug("We need more memory for the next job, so finishing")
                break
            if successorCores > job.cores:
                logger.debug("We need more cores for the next job, so finishing")
                break
            if successorsDisk > job.disk:
                logger.debug("We need more disk for the next job, so finishing")
                break
            if successorPredecessorID != None: 
                logger.debug("The job has multiple predecessors, we must return to the leader.")
                break
          
            ##########################################
            #We have a single successor job.
            #We load the successor job and transplant its command and stack
            #into the current job so that it can be run
            #as if it were a command that were part of the current job.
            #We can then delete the successor job in the jobStore, as it is
            #wholly incorporated into the current job.
            ##########################################
            
            #Remove the successor job
            job.stack.pop()
            
            #Load the successor job
            successorJob = jobStore.load(successorJobStoreID)
            #These should all match up
            assert successorJob.memory == successorMemory
            assert successorJob.cores == successorCores
            assert successorJob.predecessorsFinished == set()
            assert successorJob.predecessorNumber == 1
            assert successorJob.command != None
            assert successorJobStoreID == successorJob.jobStoreID
            
            #Transplant the command and stack to the current job
            job.command = successorJob.command
            job.stack += successorJob.stack
            assert job.memory >= successorJob.memory
            assert job.cores >= successorJob.cores
            
            #Checkpoint the job and delete the successorJob
            job.jobsToDelete = [ successorJob.jobStoreID ]
            jobStore.update(job)
            jobStore.delete(successorJob.jobStoreID)
            
            logger.debug("Starting the next job")
        
        ##########################################
        #Finish up the stats
        ##########################################
        if config.stats:
            totalCPUTime, totalMemoryUsage = getTotalCpuTimeAndMemoryUsage()
            elementNode.attrib["time"] = str(time.time() - startTime)
            elementNode.attrib["clock"] = str(totalCPUTime - startClock)
            elementNode.attrib["memory"] = str(totalMemoryUsage)
        for message in messages:
            ET.SubElement(messageNode, "message").text = message
        
        logger.info("Finished running the chain of jobs on this node, we ran for a total of %f seconds", time.time() - startTime)
    
    ##########################################
    #Trapping where worker goes wrong
    ##########################################
    except: #Case that something goes wrong in worker
        traceback.print_exc()
        logger.error("Exiting the worker because of a failed job on host %s", socket.gethostname())
        job = jobStore.load(jobStoreID)
        job.setupJobAfterFailure(config)
        workerFailed = True

    ##########################################
    #Cleanup
    ##########################################
    
    #Close the worker logging
    #Flush at the Python level
    sys.stdout.flush()
    sys.stderr.flush()
    #Flush at the OS level
    os.fsync(1)
    os.fsync(2)
    
    #Close redirected stdout and replace with the original standard output.
    os.dup2(origStdOut, 1)
    
    #Close redirected stderr and replace with the original standard error.
    os.dup2(origStdOut, 2)
    
    #sys.stdout and sys.stderr don't need to be modified at all. We don't need
    #to call redirectLoggerStreamHandlers since they still log to sys.stderr
    
    #Close our extra handles to the original standard output and standard error
    #streams, so we don't leak file handles.
    os.close(origStdOut)
    os.close(origStdErr)
    
    #Now our file handles are in exactly the state they were in before.
    
    #Copy back the log file to the global dir, if needed
    if workerFailed:
        truncateFile(tempWorkerLogPath)
        job.setLogFile(tempWorkerLogPath, jobStore)
        os.remove(tempWorkerLogPath)
        jobStore.update(job)
    elif debugging: # write log messages
        truncateFile(tempWorkerLogPath)
        with open(tempWorkerLogPath, 'r') as logFile:
            logMessages = logFile.read().splitlines()
        for logMessage in logMessages:
            ET.SubElement(messageNode, "log").text = jobStoreID+"!"+logMessage

    if (debugging or config.stats or messages) and not workerFailed: # We have stats/logging to report back
        jobStore.writeStatsAndLogging(ET.tostring(elementNode))

    #Remove the temp dir
    shutil.rmtree(localWorkerTempDir)
    
    #This must happen after the log file is done with, else there is no place to put the log
    if (not workerFailed) and job.command == None and len(job.stack) == 0:
        #We can now safely get rid of the job
        jobStore.delete(job.jobStoreID)
       
if __name__ == '__main__':
    logging.basicConfig()
    main()



#!/usr/bin/env python

# Copyright (C) 2015 UCSC Computational Genomics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import absolute_import
import unittest
import os
import tempfile
import shutil
import time
import math
import datetime

from toil.job import Job
from toil.test import ToilTest


class DependenciesTest(ToilTest):
    """
    Test program designed to catch two bugs encountered when developing progressive cactus:

    1) spawning a daemon process causes indefinite toil hangs

    2) toil does not properly parallelize jobs in certain types of recursions

    If jobs get hung up until the daemon process finsishes, that would be a case of bug 1).  If
    jobs do not get issued in parallel (ie the begin UP does not happen concurrently for the
    leaves of the comb tree), then that is a case of bug 2).  This is now verified if a log file
    is specified (--logFile [path] option)

    --Glenn Hickey
    """


    def setUp(self):
        super(DependenciesTest, self).setUp()
        self.jobStore = self._createTempDir('jobStore')
        self.tempDir = self._createTempDir('tempDir')

    # FIXME: test methods can't have parameters, AFAIK (Hannes)

    @unittest.expectedFailure
    def testDependencies(self, batchSystem="singleMachine"):
        """
        Only done in singleMachine for now.  Experts can run manually on other systems if they so choose.
        """

        def fn(tree, maxCores, maxThreads, size, coresPerJob, sleepTime):
            """
            Function runs the dependencies test
            """
            startTime = datetime.datetime.now()

            if os.path.exists(self.jobStore):
                shutil.rmtree(self.jobStore)

            logFd, logName = tempfile.mkstemp(prefix="log", dir=self.tempDir)

            # Switch out the tree with the topology
            if tree == "star":
                tree = starTree(size)
            elif tree == "balanced":
                tree = balancedTree()
            elif tree == "fly":
                tree = flyTree()
            else:
                tree = combTree(size)

            options = Job.Runner.getDefaultOptions(self.jobStore)
            options.maxThreads = maxThreads
            options.batchSystem = batchSystem
            options.logFile = logName
            options.maxCores = maxCores

            baseJob = FirstJob(tree, "Anc00", sleepTime, startTime, int(coresPerJob))
            Job.Runner.startToil(baseJob, options)
            checkLog(logName, maxCores, maxThreads, coresPerJob, sleepTime)

            os.close(logFd)

        fn("comb", 10, 100, 100, 1, 10)
        fn("comb", 200, 100, 100, 20, 10)

        fn("fly", 10, 8, 100, 1, 10)
        fn("fly", 10, 8, 100, 2, 10)

        fn("balanced", 5, 10, 100, 1, 10)
        fn("balanced", 5, 10, 100, 3, 10)


###############################################
# Classes/functions for testing dependencies
###############################################    

def writeLog(fileStore, msg, startTime):
    timeStamp = str(datetime.datetime.now() - startTime)
    fileStore.logToMaster("%s %s" % (timeStamp, msg))


def balancedTree():
    t = dict()
    t["Anc00"] = [1, 2]
    t[1] = [11, 12]
    t[2] = [21, 22]
    t[11] = [111, 112]
    t[12] = [121, 122]
    t[21] = [211, 212]
    t[22] = [221, 222]
    t[111] = [1111, 1112]
    t[112] = [1121, 1122]
    t[121] = [1211, 1212]
    t[122] = [1221, 1222]
    t[211] = [2111, 2112]
    t[212] = [2121, 2122]
    t[221] = [2211, 2212]
    t[222] = [2221, 2222]
    t[1111] = []
    t[1112] = []
    t[1121] = []
    t[1122] = []
    t[1211] = []
    t[1212] = []
    t[1221] = []
    t[1222] = []
    t[2111] = []
    t[2112] = []
    t[2121] = []
    t[2122] = []
    t[2211] = []
    t[2212] = []
    t[2221] = []
    t[2222] = []
    return t


def starTree(n=10):
    t = dict()
    t["Anc00"] = range(1, n)
    for i in range(1, n):
        t[i] = []
    return t


# odd numbers are leaves
def combTree(n=100):
    t = dict()
    for i in range(0, n):
        if i % 2 == 0:
            t[i] = [i + 1, i + 2]
        else:
            t[i] = []
        t[i + 1] = []
        t[i + 2] = []
    t["Anc00"] = t[0]
    return t


# dependencies of the internal nodes of the fly12 tree
# note that 2, 5, 8 and 10 have no dependencies
def flyTree():
    t = dict()
    t["Anc00"] = ["Anc01", "Anc03"]
    t["Anc02"] = []
    t["Anc01"] = ["Anc02"]
    t["Anc03"] = ["Anc04"]
    t["Anc04"] = ["Anc05", "Anc06"]
    t["Anc05"] = []
    t["Anc06"] = ["Anc07"]
    t["Anc07"] = ["Anc08", "Anc09"]
    t["Anc08"] = []
    t["Anc09"] = ["Anc10"]
    t["Anc10"] = []
    return t


class FirstJob(Job):
    def __init__(self, tree, event, sleepTime, startTime, cores):
        Job.__init__(self, cores=cores)
        self.tree = tree
        self.event = event
        self.sleepTime = sleepTime
        self.startTime = startTime
        self.cores = cores

    def run(self, fileStore):
        time.sleep(1)
        self.addChild(DownJob(self.tree, self.event,
                              self.sleepTime, self.startTime, self.cores))

        self.addFollowOn(LastJob())


class LastJob(Job):
    def __init__(self):
        Job.__init__(self)

    def run(self, fileStore):
        time.sleep(1)


class DownJob(Job):
    def __init__(self, tree, event, sleepTime, startTime, cores):
        Job.__init__(self, cores=cores)
        self.tree = tree
        self.event = event
        self.sleepTime = sleepTime
        self.startTime = startTime
        self.cores = cores

    def run(self, fileStore):
        writeLog(fileStore, "begin Down: %s" % self.event, self.startTime)
        children = self.tree[self.event]
        for child in children:
            writeLog(fileStore, "add %s as child of %s" % (child, self.event),
                     self.startTime)
            self.addChild(DownJob(self.tree, child,
                                  self.sleepTime, self.startTime, self.cores))

        if len(children) == 0:
            self.addFollowOn(UpJob(self.tree, self.event,
                                   self.sleepTime, self.startTime, self.cores))
        return 0


class UpJob(Job):
    def __init__(self, tree, event, sleepTime, startTime, cores):
        Job.__init__(self, cores=cores)
        self.tree = tree
        self.event = event
        self.sleepTime = sleepTime
        self.startTime = startTime
        self.cores = cores

    """
    def spawnDaemon(self, command):
        ""
        Launches a command as a daemon.  It will need to be explicitly killed
        ""
        sonLib_daemonize_py = os.path.join( os.path.dirname( os.path.abspath( __file__ ) ), 'sonLib_daemonize.py' )
        #return system( "%s \'%s\'" % ( sonLib_daemonize_py, command) )
    """

    def run(self, fileStore):
        writeLog(fileStore, "begin UP: %s" % self.event, self.startTime)

        time.sleep(self.sleepTime)
        # self.spawnDaemon("sleep %s" % str(int(self.sleepTime) * 10))
        writeLog(fileStore, "end UP: %s" % self.event, self.startTime)


# let k = maxThreads.  we make sure that jobs are fired in batches of k
# so the first k jobs all happen within epsilon time of each other,
# same for the next k jobs and so on.  we allow at most alpha time
# between the different batches (ie between k+1 and k).  
def checkLog(logFile, maxCores, maxThreads, coresPerJob, sleepTime):
    epsilon = float(sleepTime) / 2.0
    alpha = sleepTime * 2.0
    logFile = open(logFile, "r")
    stamps = []
    for logLine in logFile:
        if "begin UP" in logLine:
            chunks = logLine.split()
            assert len(chunks) == 13
            timeString = chunks[9]
            timeObj = datetime.datetime.strptime(timeString, "%H:%M:%S.%f")
            timeStamp = timeObj.hour * 3600. + timeObj.minute * 60. + \
                        timeObj.second + timeObj.microsecond / 1000000.
            stamps.append(timeStamp)

    stamps.sort()

    maxThreads = int(maxThreads)
    maxCores = int(maxCores)
    maxConcurrentJobs = min(maxThreads, maxCores)
    coresPerThread = float(maxCores) / maxConcurrentJobs
    coresPerJob = int(coresPerJob)
    assert coresPerJob >= 1
    assert coresPerThread >= 1
    threadsPerJob = 1
    if coresPerJob > coresPerThread:
        threadsPerJob = math.ceil(coresPerJob / coresPerThread)
    maxConcurrentJobs = int(maxConcurrentJobs / threadsPerJob)
    # print "Info on jobs", coresPerThread, coresPerJob, threadsPerJob, maxConcurrentJobs
    assert maxConcurrentJobs >= 1
    for i in range(1, len(stamps)):
        delta = stamps[i] - stamps[i - 1]
        if i % maxConcurrentJobs != 0:
            if delta > epsilon:
                raise RuntimeError("jobs out of sync: i=%d delta=%f threshold=%f" %
                                   (i, delta, epsilon))
        elif delta > alpha:
            raise RuntimeError("jobs out of sync: i=%d delta=%f threshold=%f" %
                               (i, delta, alpha))

    logFile.close()


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

# Copyright (C) 2015 UCSC Computational Genomics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The leader script (of the leader/worker pair) for running jobs.
"""
from __future__ import absolute_import
import logging
import sys
import os.path
import time
import xml.etree.cElementTree as ET

from toil import Process, Queue
from toil.lib.bioio import getTotalCpuTime, logStream
from toil.common import toilPackageDirPath

logger = logging.getLogger( __name__ )

####################################################
##Stats/logging aggregation
####################################################

def statsAndLoggingAggregatorProcess(jobStore, stop):
    """
    The following function is used for collating stats/reporting log messages from the workers.
    Works inside of a separate process, collates as long as the stop flag is not True.
    """
    #Overall timing
    startTime = time.time()
    startClock = getTotalCpuTime()

    #Start off the stats file
    with jobStore.writeSharedFileStream("statsAndLogging.xml") as fileHandle:
        fileHandle.write('<?xml version="1.0" ?><stats>')
        
        #Call back function
        def statsAndLoggingCallBackFn(fileHandle2):
            node = ET.parse(fileHandle2).getroot()
            nodesNamed = node.find("messages").findall
            for message in nodesNamed("message"):
                logger.warn("Got message from job at time: %s : %s",
                                    time.strftime("%m-%d-%Y %H:%M:%S"), message.text)
            for log in nodesNamed("log"):
                logger.info("%s:     %s" %
                                    tuple(log.text.split("!",1)))# the jobID is separated from log by "!"
            ET.ElementTree(node).write(fileHandle)
        
        #The main loop
        timeSinceOutFileLastFlushed = time.time()
        while True:
            if not stop.empty(): #This is a indirect way of getting a message to
                #the process to exit
                jobStore.readStatsAndLogging(statsAndLoggingCallBackFn)
                break
            if jobStore.readStatsAndLogging(statsAndLoggingCallBackFn) == 0:
                time.sleep(0.5) #Avoid cycling too fast
            if time.time() - timeSinceOutFileLastFlushed > 60: #Flush the
                #results file every minute
                fileHandle.flush()
                timeSinceOutFileLastFlushed = time.time()

        #Finish the stats file
        fileHandle.write("<total_time time='%s' clock='%s'/></stats>" % \
                         (str(time.time() - startTime), str(getTotalCpuTime() - startClock)))

####################################################
##Following encapsulates interactions with the batch system class.
####################################################

class JobBatcher:
    """
    Class works with jobBatcherWorker to submit jobs to the batch system.
    """
    def __init__(self, config, batchSystem, jobStore, toilState):
        self.config = config
        self.jobStore = jobStore
        self.jobStoreString = config.jobStore
        self.toilState = toilState
        self.jobBatchSystemIDToJobStoreIDHash = {}
        self.batchSystem = batchSystem
        self.jobsIssued = 0
        self.workerPath = os.path.join(toilPackageDirPath(), "worker.py")
        self.reissueMissingJobs_missingHash = {} #Hash to store number of observed misses

    def issueJob(self, jobStoreID, memory, cores, disk):
        """
        Add a job to the queue of jobs
        """
        self.jobsIssued += 1
        jobCommand = "%s -E %s %s %s" % (sys.executable, self.workerPath, self.jobStoreString, jobStoreID)
        jobBatchSystemID = self.batchSystem.issueBatchJob(jobCommand, memory, cores, disk)
        self.jobBatchSystemIDToJobStoreIDHash[jobBatchSystemID] = jobStoreID
        logger.debug("Issued job with job store ID: %s and job batch system ID: "
                     "%s and cores: %i, disk: %i, and memory: %i",
                     jobStoreID, str(jobBatchSystemID), cores, disk, memory)

    def issueJobs(self, jobs):
        """
        Add a list of jobs, each represented as a tuple of
        (jobStoreID, memory, cores, disk).
        """
        for jobStoreID, memory, cores, disk in jobs:
            self.issueJob(jobStoreID, memory, cores, disk)

    def getNumberOfJobsIssued(self):
        """
        Gets number of jobs that have been added by issueJob(s) and not
        removed by removeJobID
        """
        assert self.jobsIssued >= 0
        return self.jobsIssued

    def getJob(self, jobBatchSystemID):
        """
        Gets the job file associated the a given id
        """
        return self.jobBatchSystemIDToJobStoreIDHash[jobBatchSystemID]

    def hasJob(self, jobBatchSystemID):
        """
        Returns true if the jobBatchSystemID is in the list of jobs.
        """
        return self.jobBatchSystemIDToJobStoreIDHash.has_key(jobBatchSystemID)

    def getJobIDs(self):
        """
        Gets the set of jobs currently issued.
        """
        return self.jobBatchSystemIDToJobStoreIDHash.keys()

    def removeJobID(self, jobBatchSystemID):
        """
        Removes a job from the jobBatcher.
        """
        assert jobBatchSystemID in self.jobBatchSystemIDToJobStoreIDHash
        self.jobsIssued -= 1
        jobStoreID = self.jobBatchSystemIDToJobStoreIDHash.pop(jobBatchSystemID)
        return jobStoreID
    
    def killJobs(self, jobsToKill):
        """
        Kills the given set of jobs and then sends them for processing
        """
        if len(jobsToKill) > 0:
            self.batchSystem.killBatchJobs(jobsToKill)
            for jobBatchSystemID in jobsToKill:
                self.processFinishedJob(jobBatchSystemID, 1)
    
    #Following functions handle error cases for when jobs have gone awry with the batch system.

    def reissueOverLongJobs(self):
        """
        Check each issued job - if it is running for longer than desirable
        issue a kill instruction.
        Wait for the job to die then we pass the job to processFinishedJob.
        """
        maxJobDuration = self.config.maxJobDuration
        jobsToKill = []
        if maxJobDuration < 10000000:  # We won't bother doing anything if the rescue
            # time is more than 16 weeks.
            runningJobs = self.batchSystem.getRunningBatchJobIDs()
            for jobBatchSystemID in runningJobs.keys():
                if runningJobs[jobBatchSystemID] > maxJobDuration:
                    logger.warn("The job: %s has been running for: %s seconds, more than the "
                                "max job duration: %s, we'll kill it",
                                str(self.getJob(jobBatchSystemID)),
                                str(runningJobs[jobBatchSystemID]),
                                str(maxJobDuration))
                    jobsToKill.append(jobBatchSystemID)
            self.killJobs(jobsToKill)
    
    def reissueMissingJobs(self, killAfterNTimesMissing=3):
        """
        Check all the current job ids are in the list of currently running batch system jobs.
        If a job is missing, we mark it as so, if it is missing for a number of runs of
        this function (say 10).. then we try deleting the job (though its probably lost), we wait
        then we pass the job to processFinishedJob.
        """
        runningJobs = set(self.batchSystem.getIssuedBatchJobIDs())
        jobBatchSystemIDsSet = set(self.getJobIDs())
        #Clean up the reissueMissingJobs_missingHash hash, getting rid of jobs that have turned up
        missingJobIDsSet = set(self.reissueMissingJobs_missingHash.keys())
        for jobBatchSystemID in missingJobIDsSet.difference(jobBatchSystemIDsSet):
            self.reissueMissingJobs_missingHash.pop(jobBatchSystemID)
            logger.warn("Batch system id: %s is no longer missing", str(jobBatchSystemID))
        assert runningJobs.issubset(jobBatchSystemIDsSet) #Assert checks we have
        #no unexpected jobs running
        jobsToKill = []
        for jobBatchSystemID in set(jobBatchSystemIDsSet.difference(runningJobs)):
            jobStoreID = self.getJob(jobBatchSystemID)
            if self.reissueMissingJobs_missingHash.has_key(jobBatchSystemID):
                self.reissueMissingJobs_missingHash[jobBatchSystemID] = \
                self.reissueMissingJobs_missingHash[jobBatchSystemID]+1
            else:
                self.reissueMissingJobs_missingHash[jobBatchSystemID] = 1
            timesMissing = self.reissueMissingJobs_missingHash[jobBatchSystemID]
            logger.warn("Job store ID %s with batch system id %s is missing for the %i time",
                        jobStoreID, str(jobBatchSystemID), timesMissing)
            if timesMissing == killAfterNTimesMissing:
                self.reissueMissingJobs_missingHash.pop(jobBatchSystemID)
                jobsToKill.append(jobBatchSystemID)
        self.killJobs(jobsToKill)
        return len( self.reissueMissingJobs_missingHash ) == 0 #We use this to inform
        #if there are missing jobs

    def processFinishedJob(self, jobBatchSystemID, resultStatus):
        """
        Function reads a processed job file and updates it state.
        """    
        jobStoreID = self.removeJobID(jobBatchSystemID)
        if self.jobStore.exists(jobStoreID):
            job = self.jobStore.load(jobStoreID)
            if job.logJobStoreFileID is not None:
                logger.warn("The job seems to have left a log file, indicating failure: %s", jobStoreID)
                with job.getLogFileHandle( self.jobStore ) as logFileStream:
                    logStream( logFileStream, jobStoreID, logger.warn )
            assert job not in self.toilState.updatedJobs
            if resultStatus != 0:
                if job.logJobStoreFileID is None:
                    logger.warn("No log file is present, despite job failing: %s", jobStoreID)
                job.setupJobAfterFailure(self.config)
            self.toilState.updatedJobs.add(job) #Now we know the
            #job is done we can add it to the list of updated job files
            logger.debug("Added job: %s to active jobs", jobStoreID)
        else:  #The job is done
            if resultStatus != 0:
                logger.warn("Despite the batch system claiming failure the "
                            "job %s seems to have finished and been removed", jobStoreID)
            self._updatePredecessorStatus(jobStoreID)
            
    def _updatePredecessorStatus(self, jobStoreID):
        """
        Update status of a predecessor for finished successor job.
        """
        if jobStoreID not in self.toilState.successorJobStoreIDToPredecessorJobs:
            #We have reach the root job
            assert len(self.toilState.updatedJobs) == 0
            assert len(self.toilState.successorJobStoreIDToPredecessorJobs) == 0
            assert len(self.toilState.successorCounts) == 0
            return
        for predecessorJob in self.toilState.successorJobStoreIDToPredecessorJobs.pop(jobStoreID):
            self.toilState.successorCounts[predecessorJob] -= 1
            assert self.toilState.successorCounts[predecessorJob] >= 0
            if self.toilState.successorCounts[predecessorJob] == 0: #Job is done
                self.toilState.successorCounts.pop(predecessorJob)
                logger.debug("Job %s has all its successors run successfully", \
                             predecessorJob.jobStoreID)
                assert predecessorJob not in self.toilState.updatedJobs
                self.toilState.updatedJobs.add(predecessorJob) #Now we know
                #the job is done we can add it to the list of updated job files

##########################################
#Class to represent the state of the toil in memory. Loads this
#representation from the toil, in the process cleaning up
#the state of the jobs in the jobtree.
##########################################    
  
class ToilState( object ):
    """
    Represents a snapshot of the jobs in the jobStore.
    """
    def __init__( self, jobStore, rootJob ):
        # This is a hash of jobs, referenced by jobStoreID, to their predecessor jobs.
        self.successorJobStoreIDToPredecessorJobs = { }
        # Hash of jobs to counts of numbers of successors issued.
        # There are no entries for jobs
        # without successors in this map. 
        self.successorCounts = { }
        # Jobs that are ready to be processed
        self.updatedJobs = set( )
        ##Algorithm to build this information
        self._buildToilState(rootJob, jobStore)

    def _buildToilState(self, job, jobStore):
        """
        Traverses tree of jobs from the root job (rootJob) building the
        ToilState class.
        """
        if job.command != None or len(job.stack) == 0: #If the job has a command
            #or is ready to be deleted it is ready to be processed
            self.updatedJobs.add(job)
        else: #There exist successors
            self.successorCounts[job] = len(job.stack[-1])
            for successorJobStoreTuple in job.stack[-1]:
                successorJobStoreID = successorJobStoreTuple[0]
                if successorJobStoreID not in self.successorJobStoreIDToPredecessorJobs:
                    #Given that the successor job does not yet point back at a
                    #predecessor we have not yet considered it, so we call the function
                    #on the successor
                    self.successorJobStoreIDToPredecessorJobs[successorJobStoreID] = [job]
                    self._buildToilState(jobStore.load(successorJobStoreID), jobStore)
                else:
                    #We have already looked at the successor, so we don't recurse, 
                    #but we add back a predecessor link
                    self.successorJobStoreIDToPredecessorJobs[successorJobStoreID].append(job)

class FailedJobsException( Exception ):
    def __init__( self, jobStoreString, numberOfFailedJobs ):
        super( FailedJobsException, self ).__init__( "The job store '%s' contains %i failed jobs" % (jobStoreString, numberOfFailedJobs))
        self.jobStoreString = jobStoreString
        self.numberOfFailedJobs = numberOfFailedJobs
        
def mainLoop(config, batchSystem, jobStore, rootJob):
    """
    This is the main loop from which jobs are issued and processed.
    
    :raises: toil.leader.FailedJobsException if at the end of function their remain
    failed jobs
    """

    ##########################################
    #Get a snap shot of the current state of the jobs in the jobStore
    ##########################################

    toilState = ToilState(jobStore, rootJob)

    ##########################################
    #Load the jobBatcher class - used to track jobs submitted to the batch-system
    ##########################################

    #Kill any jobs on the batch system queue from the last time.
    assert len(batchSystem.getIssuedBatchJobIDs()) == 0 #Batch system must start with no active jobs!
    logger.info("Checked batch system has no running jobs and no updated jobs")

    jobBatcher = JobBatcher(config, batchSystem, jobStore, toilState)
    logger.info("Found %s jobs to start and %i jobs with successors to run",
                len(toilState.updatedJobs), len(toilState.successorCounts))

    ##########################################
    #Start the stats/logging aggregation process
    ##########################################

    stopStatsAndLoggingAggregatorProcess = Queue() #When this is s
    worker = Process(target=statsAndLoggingAggregatorProcess,
                     args=(jobStore, stopStatsAndLoggingAggregatorProcess))
    worker.start() 
    try:

        ##########################################
        #The main loop in which jobs are scheduled/processed
        ##########################################

        #Sets up the timing of the job rescuing method
        timeSinceJobsLastRescued = time.time()
        #Number of jobs that can not be completed successful after exhausting retries
        totalFailedJobs = 0
        logger.info("Starting the main loop")
        while True:

            ##########################################
            #Process jobs that are ready to be scheduled/have successors to schedule
            ##########################################

            if len(toilState.updatedJobs) > 0:
                logger.debug("Built the jobs list, currently have %i jobs to update and %i jobs issued",
                             len(toilState.updatedJobs), jobBatcher.getNumberOfJobsIssued())

                for job in toilState.updatedJobs:
                    #If the job has a command it must be run before any successors
                    if job.command != None:
                        if job.remainingRetryCount > 0:
                            jobBatcher.issueJob(job.jobStoreID, job.memory, job.cores, job.disk)
                        else:
                            totalFailedJobs += 1
                            logger.warn("Job: %s is completely failed", job.jobStoreID)

                    #There exist successors to run
                    elif len(job.stack) > 0:
                        assert len(job.stack[-1]) > 0
                        logger.debug("Job: %s has %i successors to schedule",
                                     job.jobStoreID, len(job.stack[-1]))
                        #Record the number of successors that must be completed before
                        #the job can be considered again
                        assert job not in toilState.successorCounts
                        toilState.successorCounts[job] = len(job.stack[-1])
                        #List of successors to schedule
                        successors = []
                        #For each successor schedule if all predecessors have been
                        #completed
                        for successorJobStoreID, memory, cores, disk, predecessorID in job.stack.pop():
                            #Build map from successor to predecessors.
                            if successorJobStoreID not in toilState.successorJobStoreIDToPredecessorJobs:
                                toilState.successorJobStoreIDToPredecessorJobs[successorJobStoreID] = []
                            toilState.successorJobStoreIDToPredecessorJobs[successorJobStoreID].append(job)
                            #Case that the job has multiple predecessors
                            if predecessorID != None:
                                #Load the wrapped job
                                job2 = jobStore.load(successorJobStoreID)
                                #Remove the predecessor from the list of predecessors
                                job2.predecessorsFinished.add(predecessorID)
                                #Checkpoint
                                jobStore.update(job2)
                                #If the jobs predecessors have all not all completed then
                                #ignore the job
                                assert len(job2.predecessorsFinished) >= 1
                                assert len(job2.predecessorsFinished) <= job2.predecessorNumber
                                if len(job2.predecessorsFinished) < job2.predecessorNumber:
                                    continue
                            successors.append((successorJobStoreID, memory, cores, disk))
                        jobBatcher.issueJobs(successors)

                    #There are no remaining tasks to schedule within the job, but
                    #we schedule it anyway to allow it to be deleted.

                    #TODO: An alternative would be simple delete it here and add it to the
                    #list of jobs to process, or (better) to create an asynchronous
                    #process that deletes jobs and then feeds them back into the set
                    #of jobs to be processed
                    else:
                        if job.remainingRetryCount > 0:
                            jobBatcher.issueJob(job.jobStoreID,
                                                config.defaultMemory,
                                                config.defaultCores,
                                                config.defaultDisk)
                            logger.debug("Job: %s is empty, we are scheduling to clean it up", job.jobStoreID)
                        else:
                            totalFailedJobs += 1
                            logger.warn("Job: %s is empty but completely failed - something is very wrong", job.jobStoreID)

                toilState.updatedJobs = set() #We've considered them all, so reset

            ##########################################
            #The exit criterion
            ##########################################

            if jobBatcher.getNumberOfJobsIssued() == 0:
                logger.info("Only failed jobs and their dependents (%i total) are remaining, so exiting.", totalFailedJobs)
                break

            ##########################################
            #Gather any new, updated job from the batch system
            ##########################################

            #Asks the batch system what jobs have been completed,
            #give
            updatedJob = batchSystem.getUpdatedBatchJob(10)
            if updatedJob != None:
                jobBatchSystemID, result = updatedJob
                if jobBatcher.hasJob(jobBatchSystemID):
                    if result == 0:
                        logger.debug("Batch system is reporting that the job with "
                                     "batch system ID: %s and job store ID: %s ended successfully",
                                     jobBatchSystemID, jobBatcher.getJob(jobBatchSystemID))
                    else:
                        logger.warn("Batch system is reporting that the job with "
                                    "batch system ID: %s and job store ID: %s failed with exit value %i",
                                    jobBatchSystemID, jobBatcher.getJob(jobBatchSystemID), result)
                    jobBatcher.processFinishedJob(jobBatchSystemID, result)
                else:
                    logger.warn("A result seems to already have been processed "
                                "for job with batch system ID: %i", jobBatchSystemID)
            else:
                ##########################################
                #Process jobs that have gone awry
                ##########################################

                #In the case that there is nothing happening
                #(no updated job to gather for 10 seconds)
                #check if their are any jobs that have run too long
                #(see JobBatcher.reissueOverLongJobs) or which
                #have gone missing from the batch system (see JobBatcher.reissueMissingJobs)
                if (time.time() - timeSinceJobsLastRescued >=
                    config.rescueJobsFrequency): #We only
                    #rescue jobs every N seconds, and when we have
                    #apparently exhausted the current job supply
                    jobBatcher.reissueOverLongJobs()
                    logger.info("Reissued any over long jobs")

                    hasNoMissingJobs = jobBatcher.reissueMissingJobs()
                    if hasNoMissingJobs:
                        timeSinceJobsLastRescued = time.time()
                    else:
                        timeSinceJobsLastRescued += 60 #This means we'll try again
                        #in a minute, providing things are quiet
                    logger.info("Rescued any (long) missing jobs")

        logger.info("Finished the main loop")


    finally:
        ##########################################
        #Finish up the stats/logging aggregation process
        ##########################################
        logger.info("Waiting for stats and logging collator process to finish")
        startTime = time.time()
        stopStatsAndLoggingAggregatorProcess.put(True)
        worker.join()
        logger.info("Stats/logging finished collating in %s seconds", time.time() - startTime)
        clean = config.clean
        if clean=="always" or clean == "onError" and totalFailedJobs>0 or clean == "onSuccess" and totalFailedJobs==0:
            jobStore.deleteJobStore()
    
    if totalFailedJobs > 0:
        raise FailedJobsException( config.jobStore, totalFailedJobs )



#!/usr/bin/env python

# Copyright (C) 2015 UCSC Computational Genomics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import absolute_import
from Queue import Empty


class AbstractBatchSystem:
    """An abstract (as far as python currently allows) base class
    to represent the interface the batch system must provide to the toil.
    """

    @staticmethod
    def supportsHotDeployment():
        """
        Whether this batch system supports hot deployment of the user script and toil itself. If it does,
        the __init__ method will have to accept two optional parameters in addition to the declared ones: userScript
        and toilDistribution. Both will be instances of toil.common.HotDeployedResource that represent the user
        script and a source tarball (sdist) of toil respectively.
        """
        return False

    def __init__(self, config, maxCores, maxMemory, maxDisk):
        """This method must be called.
        The config object is setup by the toilSetup script and
        has configuration parameters for the jobtree. You can add stuff
        to that script to get parameters for your batch system.
        """
        self.config = config
        self.maxCores = maxCores
        self.maxMemory = maxMemory
        self.maxDisk = maxDisk
        
    def checkResourceRequest(self, memory, cores, disk):
        """Check resource request is not greater than that available.
        """
        assert memory is not None
        assert disk is not None
        assert cores is not None
        if cores > self.maxCores:
            raise InsufficientSystemResources('cores', cores, self.maxCores)
        if memory > self.maxMemory:
            raise InsufficientSystemResources('memory', memory, self.maxMemory)
        if disk > self.maxDisk:
            raise InsufficientSystemResources('disk', disk, self.maxDisk)
        
    def issueBatchJob(self, command, memory, cores, disk):
        """Issues the following command returning a unique jobID. Command
        is the string to run, memory is an int giving
        the number of bytes the job needs to run in and cores is the number of cpu cores needed for
        the job and error-file is the path of the file to place any std-err/std-out in.
        """
        raise NotImplementedError('Abstract method: issueBatchJob')

    def killBatchJobs(self, jobIDs):
        """Kills the given job IDs.
        """
        raise NotImplementedError('Abstract method: killBatchJobs')

    # FIXME: Return value should be a set (then also fix the tests)

    def getIssuedBatchJobIDs(self):
        """A list of jobs (as jobIDs) currently issued (may be running, or maybe
        just waiting). Despite the result being a list, the ordering should not
        be depended upon.
        """
        raise NotImplementedError('Abstract method: getIssuedBatchJobIDs')
    
    def getRunningBatchJobIDs(self):
        """Gets a map of jobs (as jobIDs) currently running (not just waiting)
        and a how long they have been running for (in seconds).
        """
        raise NotImplementedError('Abstract method: getRunningBatchJobIDs')
    
    def getUpdatedBatchJob(self, maxWait):
        """Gets a job that has updated its status,
        according to the job manager. Max wait gives the number of seconds to pause
        waiting for a result. If a result is available returns (jobID, exitValue)
        else it returns None.
        """
        raise NotImplementedError('Abstract method: getUpdatedBatchJob')

    def shutdown(self):
        """Called at the completion of a toil invocation.
        Should cleanly terminate all worker threads.
        """
        raise NotImplementedError('Abstract Method: shutdown')

    @classmethod
    def getRescueBatchJobFrequency(cls):
        """Gets the period of time to wait (floating point, in seconds) between checking for 
        missing/overlong jobs.
        """
        raise NotImplementedError('Abstract method: getRescueBatchJobFrequency')

    # FIXME: Add a link to the issue tracker for this bug in multiprocessing

    # FIXME: Should be a no-op unless queue is a multiprocessing.Queue
    
    def getFromQueueSafely(self, queue, maxWait):
        """
        Returns an object from the given queue, avoiding a nasty bug in some versions of the
        multiprocessing queue python
        """
        if maxWait <= 0:
            try:
                return queue.get(block=False)
            except Empty:
                return None
        try:
            return queue.get(timeout=maxWait)
        except Empty:
            return None


class InsufficientSystemResources(Exception):
    def __init__(self, cores_or_mem, requested, available):
        self.requested = requested
        self.available = available
        self.cores_or_mem = cores_or_mem

    def __str__(self):
        return 'Requesting more {} than available. Requested: {}, Available: {}' \
               ''.format(self.cores_or_mem, self.requested, self.available)



#!/usr/bin/env python

# Copyright (C) 2015 UCSC Computational Genomics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import
import logging
import os
import subprocess
import time
import math
from Queue import Queue
from threading import Thread

from toil.batchSystems.abstractBatchSystem import AbstractBatchSystem
from toil.batchSystems.parasol import getParasolResultsFileName

logger = logging.getLogger(__name__)


class MemoryString:
    def __init__(self, string):
        if string[-1] == 'K' or string[-1] == 'M' or string[-1] == 'G':
            self.unit = string[-1]
            self.val = float(string[:-1])
        else:
            self.unit = 'B'
            self.val = float(string)
        self.bytes = self.byteVal()

    def __str__(self):
        if self.unit != 'B':
            return str(self.val) + self.unit
        else:
            return str(self.val)

    def byteVal(self):
        if self.unit == 'B':
            return self.val
        elif self.unit == 'K':
            return self.val * 1024
        elif self.unit == 'M':
            return self.val * 1048576
        elif self.unit == 'G':
            return self.val * 1073741824

    def __cmp__(self, other):
        return cmp(self.bytes, other.bytes)


def prepareQsub(cpu, mem, jobID):
    qsubline = ["qsub", "-b", "y", "-terse", "-j", "y", "-cwd", "-o", "/dev/null",
                "-e", "/dev/null", "-N", "Toil-Job_" + str(jobID)]
    try:
        path = os.environ["LD_LIBRARY_PATH"]
    except KeyError:
        pass
    else:
        qsubline.append("LD_LIBRARY_PATH=%s" % path)

    reqline = list()
    if mem is not None:
        reqline.append("vf=" + str(mem / 1024) + "K")
        reqline.append("h_vmem=" + str(mem / 1024) + "K")
    if len(reqline) > 0:
        qsubline.extend(["-hard","-l", ",".join(reqline)])
    if cpu is not None and math.ceil(cpu) > 1:
        qsubline.extend(["-pe", "smp", str(int(math.ceil(cpu)))])
    return qsubline

def qsub(qsubline):
    logger.debug("**" + " ".join(qsubline))
    process = subprocess.Popen(qsubline, stdout=subprocess.PIPE)
    result = int(process.stdout.readline().strip().split('.')[0])
    return result

def getjobexitcode(sgeJobID):
        job, task = sgeJobID
        args = ["qacct", "-j", str(job)]
        if task is not None:
             args.extend(["-t", str(task)])

        process = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        for line in process.stdout:
            if line.startswith("failed") and int(line.split()[1]) == 1:
                return 1
            elif line.startswith("exit_status"):
                logger.debug('Exit Status: {}'.format(line.split()[1]))
                return int(line.split()[1])
        return None

class Worker(Thread):
    def __init__(self, newJobsQueue, updatedJobsQueue, killQueue, killedJobsQueue, boss):
        Thread.__init__(self)
        self.newJobsQueue = newJobsQueue
        self.updatedJobsQueue = updatedJobsQueue
        self.killQueue = killQueue
        self.killedJobsQueue = killedJobsQueue
        self.waitingJobs = list()
        self.runningJobs = set()
        self.boss = boss
        self.allocatedCpus = dict()
        self.sgeJobIDs = dict()

    def getRunningJobIDs(self):
        times = {}
        currentjobs = dict((str(self.sgeJobIDs[x][0]), x) for x in self.runningJobs)
        process = subprocess.Popen(["qstat"], stdout=subprocess.PIPE)
        stdout, stderr = process.communicate()
        
        for currline in stdout.split('\n'):
            items = currline.strip().split()
            if items:
                if items[0] in currentjobs and items[4] == 'r':
                    jobstart = " ".join(items[5:7])
                    jobstart = time.mktime(time.strptime(jobstart,"%m/%d/%Y %H:%M:%S"))
                    times[currentjobs[items[0]]] = time.time() - jobstart

        return times

    def getSgeID(self, jobID):
        if not jobID in self.sgeJobIDs:
             RuntimeError("Unknown jobID, could not be converted")

        (job,task) = self.sgeJobIDs[jobID]
        if task is None:
             return str(job)
        else:
             return str(job) + "." + str(task)

    def forgetJob(self, jobID):
        self.runningJobs.remove(jobID)
        del self.allocatedCpus[jobID]
        del self.sgeJobIDs[jobID]

    def killJobs(self):
        # Load hit list:
        killList = list()
        while not self.killQueue.empty():
            killList.append(self.killQueue.get())

        # Do the dirty job
        for jobID in list(killList):
            if jobID in self.runningJobs:
                logger.debug('Killing job: {}'.format(jobID))
                process = subprocess.Popen(["qdel", self.getSgeID(jobID)])
            else:
                if jobID in self.waitingJobs:
                    self.waitingJobs.remove(jobID)
                self.killedJobsQueue.put(jobID)
                killList.remove(jobID)

        # Wait to confirm the kill
        while len(killList) > 0:
            for jobID in list(killList):
                if getjobexitcode(self.sgeJobIDs[jobID]) is not None:
                    logger.debug('Adding {} jobID to killedJobsQueue'.format(jobID))
                    self.killedJobsQueue.put(jobID)
                    killList.remove(jobID)
                    self.forgetJob(jobID)

            if len(killList) > 0:
                logger.warn("Tried to kill some jobs, but something happened and they are still going, "
                            "so I'll try again")
                time.sleep(5)

    def createJobs(self, new_job):
        # Load new job id if present:
        if new_job is not None:
            self.waitingJobs.append(new_job)

        # Launch jobs as necessary:
        while len(self.waitingJobs) > 0 and sum(self.allocatedCpus.values()) < int(self.boss.maxCores):
            jobID, cpu, memory, command = self.waitingJobs.pop(0)
            qsubline = prepareQsub(cpu, memory, jobID) + [command]
            sgeJobID = qsub(qsubline)
            self.sgeJobIDs[jobID] = (sgeJobID, None)
            self.runningJobs.add(jobID)
            self.allocatedCpus[jobID] = cpu

    def checkOnJobs(self):
        logger.debug('List of runningJobs: {}'.format(self.runningJobs))
        for jobID in list(self.runningJobs):
            exit = getjobexitcode(self.sgeJobIDs[jobID])
            if exit is not None:
                self.updatedJobsQueue.put((jobID, exit))
                self.forgetJob(jobID)
    
    def run(self):
        while True:
            new_job = None
            if not self.newJobsQueue.empty():
                new_job = self.newJobsQueue.get()
                if new_job is None:
                    logger.debug('Received queue sentinel.')
                    break
            self.killJobs()
            self.createJobs(new_job)
            self.checkOnJobs()
            time.sleep(10)

class GridengineBatchSystem(AbstractBatchSystem):
    """The interface for gridengine.
    """
    
    def __init__(self, config, maxCores, maxMemory, maxDisk):
        AbstractBatchSystem.__init__(self, config, maxCores, maxMemory, maxDisk)
        self.gridengineResultsFile = getParasolResultsFileName(config.jobStore)
        #Reset the job queue and results (initially, we do this again once we've killed the jobs)
        self.gridengineResultsFileHandle = open(self.gridengineResultsFile, 'w')
        self.gridengineResultsFileHandle.close() #We lose any previous state in this file, and ensure the files existence
        self.currentjobs = set()
        self.obtainSystemConstants()
        self.nextJobID = 0
        self.newJobsQueue = Queue()
        self.updatedJobsQueue = Queue()
        self.killQueue = Queue()
        self.killedJobsQueue = Queue()
        self.worker = Worker(self.newJobsQueue, self.updatedJobsQueue, self.killQueue, self.killedJobsQueue, self)
        self.worker.start()
        
    def __des__(self):
        #Closes the file handle associated with the results file.
        self.gridengineResultsFileHandle.close() #Close the results file, cos were done.

    def issueBatchJob(self, command, memory, cores, disk):
        self.checkResourceRequest(memory, cores, disk)
        jobID = self.nextJobID
        self.nextJobID += 1
        self.currentjobs.add(jobID)
        self.newJobsQueue.put((jobID, cores, memory, command))
        logger.debug("Issued the job command: %s with job id: %s " % (command, str(jobID)))
        return jobID

    def killBatchJobs(self, jobIDs):
        """Kills the given jobs, represented as Job ids, then checks they are dead by checking
        they are not in the list of issued jobs.
        """
        for jobID in jobIDs:
            self.killQueue.put(jobID)

        logger.debug('Jobs to be killed: {}'.format(jobIDs))
        killList = set(jobIDs)
        while len(killList) > 0:
            i = self.killedJobsQueue.get()
            if i is not None:
                killList.remove(i)
                if i in self.currentjobs:
                    self.currentjobs.remove(i)
            else:
                break

            if len(killList) > 0:
                time.sleep(5)

    def getIssuedBatchJobIDs(self):
        """Gets the list of jobs issued to SGE.
        """
        return list(self.currentjobs)
    
    def getRunningBatchJobIDs(self):
        return self.worker.getRunningJobIDs()
    
    def getUpdatedBatchJob(self, maxWait):
        i = self.updatedJobsQueue.get()
        logger.debug('UpdatedJobsQueue Item: {}'.format(i))
        if i == None:
            return None
        jobID, retcode = i
        self.updatedJobsQueue.task_done()
        self.currentjobs.remove(jobID)
        return i

    def shutdown(self):
        """
        Signals worker to shutdown (via sentinel) then cleanly joins the thread
        """
        newJobsQueue = self.newJobsQueue
        self.newJobsQueue = None

        newJobsQueue.put(None)
        self.worker.join()

    def getWaitDuration(self):
        """We give parasol a second to catch its breath (in seconds)
        """
        return 0.0

    @classmethod
    def getRescueBatchJobFrequency(cls):
        """Parasol leaks jobs, but rescuing jobs involves calls to parasol list jobs and pstat2,
        making it expensive. We allow this every 10 minutes..
        """
        return 1800 #Half an hour

    def obtainSystemConstants(self):
        p = subprocess.Popen(["qhost"], stdout = subprocess.PIPE,stderr = subprocess.STDOUT)

        line = p.stdout.readline()
        items = line.strip().split()
        num_columns = len(items)
        cpu_index = None
        mem_index = None        
        for i in range(num_columns):
                if items[i] == 'NCPU':
                        cpu_index = i
                elif items[i] == 'MEMTOT':
                        mem_index = i

        if cpu_index is None or mem_index is None:
                RuntimeError("qhost command does not return NCPU or MEMTOT columns")

        p.stdout.readline()

        self.maxCPU = 0
        self.maxMEM = MemoryString("0")
        for line in p.stdout:
                items = line.strip().split()
                if len(items) < num_columns:
                        RuntimeError("qhost output has a varying number of columns")
                if items[cpu_index] != '-' and items[cpu_index] > self.maxCPU:
                        self.maxCPU = items[cpu_index]
                if items[mem_index] != '-' and MemoryString(items[mem_index]) > self.maxMEM:
                        self.maxMEM = MemoryString(items[mem_index])

        if self.maxCPU is 0 or self.maxMEM is 0:
                RuntimeError("qhost returns null NCPU or MEMTOT info")
                
        
def main():
    pass

def _test():
    import doctest      
    return doctest.testmod()

if __name__ == '__main__':
    _test()
    main()



#!/usr/bin/env python2.7

from __future__ import absolute_import, print_function
import os
import sys

def main():
    modules = loadModules()
    try:
        command = sys.argv[1]
    except IndexError:
        printHelp(modules)
    else:
        if command == '--help':
            printHelp(modules)
        else:
            try:
                module = modules[command]
            except KeyError:
                print("Unknown option '%s'. "
                      "Pass --help to display usage information.\n" % command, file=sys.stderr)
                sys.exit(1)
            else:
                del sys.argv[1]
                module.main()


def loadModules():
    # noinspection PyUnresolvedReferences
    from toil.utils import toilKill, toilRestart, toilStats, toilStatus, toilClean
    return {name[4:].lower(): module for name, module in locals().iteritems()}


def printHelp(modules):
    usage = ("\n"
             "Usage: {name} COMMAND ...\n"
             "       {name} --help\n"
             "       {name} COMMAND --help\n\n"
             "where COMMAND is one of the following:\n\n{descriptions}\n\n")
    print(usage.format(
        name=os.path.basename(sys.argv[0]),
        commands='|'.join(modules.iterkeys()),
        descriptions='\n'.join("%s - %s" % (n, m.__doc__.strip()) for n, m in modules.iteritems())))


if __name__ == '__main__':
    main()



# Copyright (C) 2015 UCSC Computational Genomics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from contextlib import contextmanager
import logging
import os
import sys
import cPickle
from argparse import ArgumentParser

from toil.lib.bioio import addLoggingOptions, getLogLevelString, absSymPath
from toil.batchSystems.parasol import ParasolBatchSystem
from toil.batchSystems.gridengine import GridengineBatchSystem
from toil.batchSystems.singleMachine import SingleMachineBatchSystem
from toil.batchSystems.lsf import LSFBatchSystem

logger = logging.getLogger( __name__ )

def toilPackageDirPath():
    """
    Returns the absolute path of the directory that corresponds to the top-level toil package. The return value is
    guaranteed to end in '/toil'.
    """
    import toil.jobWrapper

    result = os.path.dirname(absSymPath(toil.jobWrapper.__file__))
    assert result.endswith('/toil')
    return result

class Config(object):
    """
    Class to represent configuration operations for a toil workflow run. 
    """
    def __init__(self):
        #Core options
        self.jobStore = os.path.abspath("./toil")
        self.logLevel = getLogLevelString()
        self.workDir = None
        self.stats = False

        # Because the stats option needs the jobStore to persist past the end of the run,
        # the clean default value depends the specified stats option and is determined in setOptions
        self.clean = None
        
        #Restarting the workflow options
        self.restart = False
        
        #Batch system options
        self.batchSystem = "singleMachine"
        self.scale = 1
        self.masterIP = '127.0.0.1:5050'
        self.parasolCommand = "parasol"
        
        #Resource requirements
        self.defaultMemory = 2147483648
        self.defaultCores = 1
        self.defaultDisk = 2147483648
        self.maxCores = sys.maxint
        self.maxMemory = sys.maxint
        self.maxDisk = sys.maxint
        
        #Retrying/rescuing jobs
        self.retryCount = 0
        self.maxJobDuration = sys.maxint
        self.rescueJobsFrequency = 3600
        
        #Misc
        self.maxLogFileSize=50120
        self.sseKey = None
        
    def setOptions(self, options):
        """
        Creates a config object from the options object.
        """
        from bd2k.util.humanize import human2bytes #This import is used to convert
        #from human readable quantites to integers 
        def setOption(varName, parsingFn=None, checkFn=None):
            #If options object has the option "varName" specified
            #then set the "varName" attrib to this value in the config object
            x = getattr(options, varName, None)
            if x != None:
                if parsingFn != None:
                    x = parsingFn(x)
                if checkFn != None:
                    try:
                        checkFn(x)
                    except AssertionError:
                        raise RuntimeError("The %s option has an invalid value: %s" 
                                           % (varName, x))
                setattr(self, varName, x)
            
        h2b = lambda x : human2bytes(str(x)) #Function to parse integer from string expressed in different formats
        
        def iC(minValue, maxValue=sys.maxint):
            #Returns function to check the a parameter is in a valid range
            def f(x):
                assert x >= minValue and x < maxValue
            return f
        
        #Core options
        setOption("jobStore", parsingFn=lambda x : os.path.abspath(x) 
                  if options.jobStore.startswith('.') else x)
        #TODO: LOG LEVEL STRING
        setOption("workDir")
        setOption("stats")
        setOption("clean")
        if self.stats:
            if self.clean != "never" and self.clean is not None:
                raise RuntimeError("Contradicting options passed: Clean flag is set to %s "
                                   "despite the stats flag requiring "
                                   "the jobStore to be intact at the end of the run. " 
                                   "Set clean to \'never\'" % self.clean)
            self.clean = "never"
        elif self.clean is None:
            self.clean = "onSuccess"
        
        #Restarting the workflow options
        setOption("restart") 
        
        #Batch system options
        setOption("batchSystem")
        setOption("scale", float) 
        setOption("masterIP") 
        setOption("parasolCommand")
        
        #Resource requirements
        setOption("defaultMemory", h2b, iC(1))
        setOption("defaultCores", h2b, iC(1))
        setOption("defaultDisk", h2b, iC(1))
        setOption("maxCores", h2b, iC(1))
        setOption("maxMemory", h2b, iC(1))
        setOption("maxDisk", h2b, iC(1))
        
        #Retrying/rescuing jobs
        setOption("retryCount", int, iC(0))
        setOption("maxJobDuration", int, iC(1))
        setOption("rescueJobsFrequency", int, iC(1))
        
        #Misc
        setOption("maxLogFileSize", h2b, iC(1))
        def checkSse(sseKey):
            with open(sseKey) as f:
                assert(len(f.readline().rstrip()) == 32)
        setOption("sseKey", checkFn=checkSse)

def _addOptions(addGroupFn, config):
    #
    #Core options
    #
    addOptionFn = addGroupFn("toil core options", "Options to specify the \
    location of the toil and turn on stats collation about the performance of jobs.")
    #TODO - specify how this works when path is AWS
    addOptionFn('jobStore', type=str,
                      help=("Store in which to place job management files \
                      and the global accessed temporary files"
                      "(If this is a file path this needs to be globally accessible "
                      "by all machines running jobs).\n"
                      "If the store already exists and restart is false an"
                      " ExistingJobStoreException exception will be thrown."))
    addOptionFn("--workDir", dest="workDir", default=None,
                help="Absolute path to directory where temporary files generated during the Toil run should be placed. "
                     "Default is determined by environmental variables (TMPDIR, TEMP, TMP) via mkdtemp")
    addOptionFn("--stats", dest="stats", action="store_true", default=None,
                      help="Records statistics about the toil workflow to be used by 'toil stats'.")
    addOptionFn("--clean", dest="clean", choices=['always', 'onError','never', 'onSuccess'], default=None,
                      help=("Determines the deletion of the jobStore upon completion of the program. "
                            "Choices: 'always', 'onError','never', 'onSuccess'. The --stats option requires "
                            "information from the jobStore upon completion so the jobStore will never be deleted with"
                            "that flag. If you wish to be able to restart the run, choose \'never\' or \'onSuccess\'. "
                            "Default is \'never\' if stats is enabled, and \'onSuccess\' otherwise"))

    #
    #Restarting the workflow options
    #
    addOptionFn = addGroupFn("toil options for restarting an existing workflow",
                             "Allows the restart of an existing workflow")
    addOptionFn("--restart", dest="restart", default=None, action="store_true",
                help="If --restart is specified then will attempt to restart existing workflow " 
                "at the location pointed to by the --jobStore option. Will raise an exception if the workflow does not exist")
    
    #
    #Batch system options
    #
    addOptionFn = addGroupFn("toil options for specifying the batch system",
                             "Allows the specification of the batch system, and arguments to the batch system/big batch system (see below).")
    addOptionFn("--batchSystem", dest="batchSystem", default=None,
                      help=("The type of batch system to run the job(s) with, currently can be one "
                            "of singleMachine, parasol, gridEngine, lsf or mesos'. default=%s" % config.batchSystem))
    #TODO - what is this?
    addOptionFn("--scale", dest="scale", default=None,
                help=("A scaling factor to change the value of all submitted tasks's submitted cores. "
                      "Used in singleMachine batch system. default=%s" % config.scale))
    addOptionFn("--masterIP", dest="masterIP", default=None,
                help=("The master node's ip and port number. Used in mesos batch system. default=%s" % config.masterIP))
    addOptionFn("--parasolCommand", dest="parasolCommand", default=None,
                      help="The command to run the parasol program default=%s" % config.parasolCommand)

    #
    #Resource requirements
    #
    addOptionFn = addGroupFn("toil options for cores/memory requirements",
                             "The options to specify default cores/memory requirements (if not specified by the jobs themselves), and to limit the total amount of memory/cores requested from the batch system.")
    addOptionFn("--defaultMemory", dest="defaultMemory", default=None,
                      help=("The default amount of memory to request for a job (in bytes), "
                            "by default is 2^31 = 2 gigabytes, default=%s" % config.defaultMemory))
    addOptionFn("--defaultCores", dest="defaultCores", default=None,
                      help="The number of cpu cores to dedicate a job. default=%s" % config.defaultCores)
    addOptionFn("--defaultDisk", dest="defaultDisk", default=None,
                      help="The amount of disk space to dedicate a job (in bytes). default=%s" % config.defaultDisk)
    addOptionFn("--maxCores", dest="maxCores", default=None,
                      help=("The maximum number of cpu cores to request from the batch system at any "
                            "one time. default=%s" % config.maxCores))
    addOptionFn("--maxMemory", dest="maxMemory", default=None,
                      help=("The maximum amount of memory to request from the batch \
                      system at any one time. default=%s" % config.maxMemory))
    addOptionFn("--maxDisk", dest="maxDisk", default=None,
                      help=("The maximum amount of disk space to request from the batch \
                      system at any one time. default=%s" % config.maxDisk))

    #
    #Retrying/rescuing jobs
    #
    addOptionFn = addGroupFn("toil options for rescuing/killing/restarting jobs", \
            "The options for jobs that either run too long/fail or get lost \
            (some batch systems have issues!)")
    addOptionFn("--retryCount", dest="retryCount", default=None,
                      help=("Number of times to retry a failing job before giving up and "
                            "labeling job failed. default=%s" % config.retryCount))
    addOptionFn("--maxJobDuration", dest="maxJobDuration", default=None,
                      help=("Maximum runtime of a job (in seconds) before we kill it "
                            "(this is a lower bound, and the actual time before killing "
                            "the job may be longer). default=%s" % config.maxJobDuration))
    addOptionFn("--rescueJobsFrequency", dest="rescueJobsFrequency", default=None,
                      help=("Period of time to wait (in seconds) between checking for "
                            "missing/overlong jobs, that is jobs which get lost by the batch system. Expert parameter. default=%s" % config.rescueJobsFrequency))
    
    #
    #Misc options
    #
    addOptionFn = addGroupFn("toil miscellaneous options", "Miscellaneous options")
    addOptionFn("--maxLogFileSize", dest="maxLogFileSize", default=None,
                      help=("The maximum size of a job log file to keep (in bytes), log files larger "
                            "than this will be truncated to the last X bytes. Default is 50 "
                            "kilobytes, default=%s" % config.maxLogFileSize))
    
    addOptionFn("--sseKey", dest="sseKey", default=None,
            help="Path to file containing 32 character key to be used for server-side encryption on awsJobStore. SSE will "
                 "not be used if this flag is not passed.")

def addOptions(parser, config=Config()):
    """
    Adds toil options to a parser object, either optparse or argparse.
    """
    # Wrapper function that allows toil to be used with both the optparse and
    # argparse option parsing modules
    addLoggingOptions(parser) # This adds the logging stuff.
    if isinstance(parser, ArgumentParser):
        def addGroup(headingString, bodyString):
            return parser.add_argument_group(headingString, bodyString).add_argument
        _addOptions(addGroup, config)
    else:
        raise RuntimeError("Unanticipated class passed to addOptions(), %s. Expecting "
                           "Either optparse.OptionParser or argparse.ArgumentParser" % parser.__class__)

def loadBatchSystem(config):
    """
    Load the configured batch system class, instantiate it and return the instance.
    """
    batchSystemClass, kwargs = loadBatchSystemClass(config)
    return createBatchSystem(config, batchSystemClass, kwargs)

def loadBatchSystemClass(config):
    """
    Returns a pair containing the concrete batch system class and a dictionary of keyword arguments to be passed to
    the constructor of that class.

    :param config: the current configuration
    :param key: the name of the configuration attribute that holds the configured batch system name
    """
    batchSystemName = config.batchSystem
    kwargs = dict(config=config,
                  maxCores=config.maxCores,
                  maxMemory=config.maxMemory,
                  maxDisk=config.maxDisk)
    if batchSystemName == 'parasol':
        batchSystemClass = ParasolBatchSystem
        logger.info('Using the parasol batch system')
    elif batchSystemName == 'single_machine' or batchSystemName == 'singleMachine':
        batchSystemClass = SingleMachineBatchSystem
        logger.info('Using the single machine batch system')
    elif batchSystemName == 'gridengine' or batchSystemName == 'gridEngine':
        batchSystemClass = GridengineBatchSystem
        logger.info('Using the grid engine machine batch system')
    elif batchSystemName == 'lsf' or batchSystemName == 'LSF':
        batchSystemClass = LSFBatchSystem
        logger.info('Using the lsf batch system')
    elif batchSystemName == 'mesos' or batchSystemName == 'Mesos':
        from toil.batchSystems.mesos.batchSystem import MesosBatchSystem
        batchSystemClass = MesosBatchSystem
        kwargs["masterIP"] = config.masterIP
        logger.info('Using the mesos batch system')
    else:
        raise RuntimeError('Unrecognised batch system: %s' % batchSystemName)
    return batchSystemClass, kwargs

def createBatchSystem(config, batchSystemClass, kwargs):
    """
    Returns an instance of the given batch system class, or if a big batch system is configured, a batch system
    instance that combines the given class with the configured big batch system.

    :param config: the current configuration
    :param batchSystemClass: the class to be instantiated
    :param kwargs: a list of keyword arguments to be passed to the given class' constructor
    """
    batchSystem = batchSystemClass(**kwargs)
    return batchSystem

def loadJobStore( jobStoreString, config=None ):
    """
    Loads a jobStore.

    :param jobStoreString: see exception message below
    :param config: see AbstractJobStore.__init__
    :return: an instance of a concrete subclass of AbstractJobStore
    :rtype : jobStores.abstractJobStore.AbstractJobStore
    """
    if jobStoreString[ 0 ] in '/.':
        jobStoreString = 'file:' + jobStoreString

    try:
        jobStoreName, jobStoreArgs = jobStoreString.split( ':', 1 )
    except ValueError:
        raise RuntimeError(
            'Job store string must either be a path starting in . or / or a contain at least one '
            'colon separating the name of the job store implementation from an initialization '
            'string specific to that job store. If a path starting in . or / is passed, the file '
            'job store will be used for backwards compatibility.' )

    if jobStoreName == 'file':
        from toil.jobStores.fileJobStore import FileJobStore
        return FileJobStore( jobStoreArgs, config=config )
    elif jobStoreName == 'aws':
        from toil.jobStores.awsJobStore import AWSJobStore
        region, namePrefix = jobStoreArgs.split( ':', 1 )
        return AWSJobStore( region, namePrefix, config=config )
    elif jobStoreName == 'azure':
        from toil.jobStores.azureJobStore import AzureJobStore
        account, namePrefix = jobStoreArgs.split( ':', 1 )
        return AzureJobStore( account, namePrefix, config=config )
    else:
        raise RuntimeError( "Unknown job store implementation '%s'" % jobStoreName )

def serialiseEnvironment(jobStore):
    """
    Puts the environment in a globally accessible pickle file.
    """
    #Dump out the environment of this process in the environment pickle file.
    with jobStore.writeSharedFileStream("environment.pickle") as fileHandle:
        cPickle.dump(os.environ, fileHandle)
    logger.info("Written the environment for the jobs to the environment file")

@contextmanager
def setupToil(options, userScript=None):
    """
    Creates the data-structures needed for running a toil.

    :type userScript: toil.resource.ModuleDescriptor
    """
    #Make the default config object
    config = Config()
    #Get options specified by the user
    config.setOptions(options)
    if not options.restart: #Create for the first time
        batchSystemClass, kwargs = loadBatchSystemClass(config)
        #Load the jobStore
        jobStore = loadJobStore(config.jobStore, config=config)
    else:
        #Reload the workflow
        jobStore = loadJobStore(config.jobStore)
        config = jobStore.config
        #Update the earlier config with any options that have been set
        config.setOptions(options)
        #Write these new options back to disk
        jobStore.writeConfigToStore()
        #Get the batch system class
        batchSystemClass, kwargs = loadBatchSystemClass(config)
    if (userScript is not None
        and not userScript.belongsToToil
        and batchSystemClass.supportsHotDeployment()):
        kwargs['userScript'] = userScript.saveAsResourceTo(jobStore)
        # TODO: toil distribution

    batchSystem = createBatchSystem(config, batchSystemClass, kwargs)
    try:
        serialiseEnvironment(jobStore)
        yield (config, batchSystem, jobStore)
    finally:
        logger.debug('Shutting down batch system')
        batchSystem.shutdown()



# Copyright (C) 2015 UCSC Computational Genomics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import

import sys

from mock import MagicMock, patch

from toil.resource import ModuleDescriptor, Resource
from toil.test import ToilTest


class ResourceTest( ToilTest ):
    def test( self ):
        # Create a ModuleDescriptor for the module containing ModuleDescriptor, i.e. toil.resource
        module_name = ModuleDescriptor.__module__
        self.assertEquals( module_name, 'toil.resource' )
        module = ModuleDescriptor.forModule( module_name )

        # Assert basic attributes and properties
        self.assertTrue( module.belongsToToil )
        self.assertEquals( module.name, module_name )
        self.assertEquals( module.filePath, sys.modules[ module_name ].__file__ )
        self.assertTrue( module.filePath.startswith( module.dirPath ) )
        self.assertTrue( module.dirPath.endswith( '/src' ) )

        # Before the module is saved as a resource, localize() and globalize() are identity
        # methods. This should log warnings.
        self.assertIs( module.localize( ), module )
        self.assertIs( module.globalize( ), module )

        # Create a mock job store ...
        jobStore = MagicMock( )
        # ... to generate a fake URL for the resource ...
        url = 'file://foo.zip'
        jobStore.getSharedPublicUrl.return_value = url
        # ... and save the resource to it.
        resource = module.saveAsResourceTo( jobStore )
        # Ensure that the URL generation method is actually called, ...
        jobStore.getSharedPublicUrl.assert_called_once_with( resource.pathHash )
        # ... and that ensure that writeSharedFileStream is called.
        jobStore.writeSharedFileStream.assert_called_once_with( resource.pathHash,
                                                                isProtected=False )

        # Now it gets a bit complicated: Ensure that the context manager returned by the
        # jobStore's writeSharedFileStream() method is entered and that the file handle yielded
        # by the context manager is written to once with the zipped source tree from which
        # 'toil.resource' was orginally imported. Keep the zipped tree around such that we can
        # mock the download later.
        file_handle = jobStore.writeSharedFileStream.return_value.__enter__.return_value
        # The first 0 index selects the first call of write(), the second 0 selects positional
        # instead of keyword arguments, and the third 0 selects the first positional, i.e. the
        # contents. This is a bit brittle since it assumes that all the data is written in a
        # single call to write(). If more calls are made we can easily concatenate them.
        zipFile = file_handle.write.call_args_list[ 0 ][ 0 ][ 0 ]
        self.assertTrue( zipFile.startswith( 'PK' ) ) # the magic header for ZIP files
        self.assertEquals( resource.url, url )

        # Now we're on the worker. Prepare the storage for localized resources
        Resource.prepareSystem( )
        # Register the resource for subsequent lookup.
        resource.register( )
        # Lookup the resource and ensure that the result is equal to but not the same as the
        # original resource. Lookup will also be used when we localize the module that was
        # originally used to create the resource.
        localResource = Resource.lookup( module._resourcePath )
        self.assertEquals( resource, localResource )
        self.assertIsNot( resource, localResource )

        # Now show that we can localize the module using the registered resource. Set up a mock
        # urlopen() that yields the zipped tree ...
        mock_urlopen = MagicMock( )
        mock_urlopen.return_value.read.return_value = zipFile
        with patch( 'toil.resource.urlopen', mock_urlopen ):
            # ... and use it to download and unpack the resource
            localModule = module.localize( )
        # Name and extension should be equal between original and localized resource ...
        self.assertEquals( module.name, localModule.name )
        self.assertEquals( module.extension, localModule.extension )
        # ... but the directory should be different.
        self.assertNotEquals( module.dirPath, localModule.dirPath )

        # Show that we can 'undo' localization. This is necessary when the user script's jobs are
        # invoked on the worker where they generate more child jobs.
        self.assertEquals( localModule.globalize( ), module )



#!/usr/bin/env python

# Copyright (C) 2015 UCSC Computational Genomics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
import os
import sys
import threading
import pickle
import logging
import subprocess
import traceback
from time import sleep

import psutil
import mesos.interface
from mesos.interface import mesos_pb2
import mesos.native
from toil.resource import Resource

log = logging.getLogger(__name__)


class MesosExecutor(mesos.interface.Executor):
    """
    Part of mesos framework, runs on mesos slave. A toil job is passed to it via the task.data field, and launched
    via call(toil.command). Uses the ExecutorDriver to communicate.
    """

    def __init__(self):
        super(MesosExecutor, self).__init__()
        self.popenLock = threading.Lock()
        self.runningTasks = {}
        Resource.prepareSystem()
        # FIXME: clean up resource root dir


    def registered(self, driver, executorInfo, frameworkInfo, slaveInfo):
        """
        Invoked once the executor driver has been able to successfully connect with Mesos.
        """
        log.info("Registered with framework")
        statThread = threading.Thread(target=self._sendStats, args=[driver])
        statThread.setDaemon(True)
        statThread.start()

    def reregistered(self, driver, slaveInfo):
        """
        Invoked when the executor re-registers with a restarted slave.
        """
        log.info("Re-registered")

    def disconnected(self, driver):
        """
        Invoked when the executor becomes "disconnected" from the slave (e.g., the slave is being restarted due to an upgrade).
        """
        log.critical("Disconnected from slave")

    def killTask(self, driver, taskId):
        if taskId in self.runningTasks:
            os.kill(self.runningTasks[taskId], 9)

    def shutdown(self, driver):
        log.critical("Shutting down executor...")
        for taskId, pid in self.runningTasks.items():
            self.killTask(driver, taskId)
        Resource.cleanSystem()
        log.critical("Executor shut down")

    def error(self, driver, message):
        """
        Invoked when a fatal error has occurred with the executor and/or executor driver.
        """
        log.critical("FATAL ERROR: " + message)

    def _sendStats(self, driver):
        while True:
            coresUsage = str(psutil.cpu_percent())
            ramUsage = str(psutil.virtual_memory().percent)
            driver.sendFrameworkMessage("CPU usage: %s, memory usage: %s" % (coresUsage, ramUsage))
            log.debug("Sent stats message")
            sleep(30)

    def launchTask(self, driver, task):
        """
        Invoked by SchedulerDriver when a Mesos task should be launched by this executor
        """

        def runTask():
            log.debug("Running task %s", task.task_id.value)
            sendUpdate(mesos_pb2.TASK_RUNNING)
            popen = runJob(pickle.loads(task.data))
            self.runningTasks[task.task_id.value] = popen.pid
            try:
                exitStatus = popen.wait()
                if 0 == exitStatus:
                    sendUpdate(mesos_pb2.TASK_FINISHED)
                elif -9 == exitStatus:
                    sendUpdate(mesos_pb2.TASK_KILLED)
                else:
                    sendUpdate(mesos_pb2.TASK_FAILED)
            except:
                exc_type, exc_value, exc_trace = sys.exc_info()
                sendUpdate(mesos_pb2.TASK_FAILED, message=str(traceback.format_exception_only(exc_type, exc_value)))
            finally:
                del self.runningTasks[task.task_id.value]

        def runJob(job):
            """
            :type job: toil.batchSystems.mesos.ToilJob

            :rtype: subprocess.Popen
            """
            if job.userScript:
                job.userScript.register()
            log.debug("Invoking command: '%s'", job.command)
            with self.popenLock:
                return subprocess.Popen(job.command, shell=True)

        def sendUpdate(taskState, message=''):
            log.debug("Sending status update ...")
            status = mesos_pb2.TaskStatus()
            status.task_id.value = task.task_id.value
            status.message = message
            status.state = taskState
            driver.sendStatusUpdate(status)
            log.debug("Sent status update")

        thread = threading.Thread(target=runTask)
        thread.start()

    def frameworkMessage(self, driver, message):
        """
        Invoked when a framework message has arrived for this executor.
        """
        log.debug("Received message from framework: {}".format(message))


def main(executorClass=MesosExecutor):
    logging.basicConfig(level=logging.DEBUG)
    log.debug("Starting executor")
    executor = executorClass()
    driver = mesos.native.MesosExecutorDriver(executor)
    exit_value = 0 if driver.run() == mesos_pb2.DRIVER_STOPPED else 1
    assert len(executor.runningTasks) == 0
    sys.exit(exit_value)


if __name__ == "__main__":
    main()
